from sympy import Matrix, Symbol, conjugate, sqrt

requirements_dct = {0: ['A','A'],
                    1: ['A','B'],
                    2: ['E','A'],
                    3: ['E','B'],
                    4: ['E','E'],
                    5: ['B','B'],
                    6: ['A'],
                    7: ['B'],
                    8: ['E']}

#Structure of matrix_dct values: [original H ,symmetrized H, transformation U]
matrix_dct = {0: [Matrix([[0,'A_alpha__A_beta'],
                          ['A_beta__A_alpha',0]]),

                  Matrix([[0,'A_alpha__A_beta'],
                          ['A_alpha__A_beta',0]]),

                  Matrix([[1,0],
                          [0,1]])],

              1: [Matrix([[0,'A__B'],
                          ['B__A',0]]),

                  Matrix([[0,'A__B'],
                          ['A__B',0]]),

                  Matrix([[1,0],
                          [0,1]])],

              2: [Matrix([[0,0,Symbol('+__A')],
                          [0,0,Symbol('-__A')],
                          [Symbol('A__+'),Symbol('A__-'),0]]),

                  Matrix([[0,0,Symbol('+__A')],
                          [0,0,conjugate(Symbol('+__A'))],
                          [conjugate(Symbol('+__A')),Symbol('+__A'),0]]),

                  Matrix([[1,1,0],
                          [1j,-1j,0],
                          [0,0,sqrt(2)]])*1/sqrt(2)],

              3: [Matrix([[0,0,Symbol('+__B')],
                          [0,0,Symbol('-__B')],
                          [Symbol('B__+'),Symbol('B__-'),0]]),

                  Matrix([[0,0,Symbol('+__B')],
                          [0,0,conjugate(Symbol('+__B'))],
                          [conjugate(Symbol('+__B')),Symbol('+__B'),0]]),

                  Matrix([[1,1,0],
                          [1j,-1j,0],
                          [0,0,sqrt(2)]])*1/sqrt(2)],

              4: [Matrix([[0,0,Symbol('+_alpha__+_beta'),Symbol('+_alpha__-_beta')],
                          [0,0,Symbol('-_alpha__+_beta'),Symbol('-_alpha__-_beta')],
                          [Symbol('+_beta__+_alpha'),Symbol('+_beta__-_alpha'),0,0],
                          [Symbol('-_beta__+_alpha'),Symbol('-_beta__-_alpha'),0,0]]),

                  Matrix([[0,0,Symbol('+_alpha__+_beta'),Symbol('+_alpha__-_beta')],
                          [0,0,conjugate(Symbol('+_alpha__-_beta')),conjugate(Symbol('+_alpha__+_beta'))],
                          [conjugate(Symbol('+_alpha__+_beta')),Symbol('+_alpha__-_beta'),0,0],
                          [conjugate(Symbol('+_alpha__-_beta')),Symbol('+_alpha__+_beta'),0,0]]),

                  Matrix([[1,1,0,0],
                          [1j,-1j,0,0],
                          [0,0,1,1],
                          [0,0,1j,-1j]])*1/sqrt(2)],

              5: [Matrix([[0,'B_alpha__B_beta'],
                          ['B_beta__B_alpha',0]]),

                  Matrix([[0,'B_alpha__B_beta'],
                          ['B_alpha__B_beta',0]]),

                  Matrix([[1,0],
                          [0,1]])],

              6: [Matrix([['A__A']]),

                  Matrix([['A__A']]),

                  Matrix([[1]])],

              7: [Matrix([['B__B']]),

                  Matrix([['B__B']]),

                  Matrix([[1]])],

              8: [Matrix([[Symbol('+__+'), Symbol('+__-')],
                          [Symbol('-__+'), Symbol('-__-')]]),

                  Matrix([[Symbol('+__+'), Symbol('+__-')],
                          [conjugate(Symbol('+__-')), Symbol('+__+')]]),

                  Matrix([[1,1],
                          [1j,-1j]])*1/sqrt(2)]

              }

def return_matrices(symmetry,states):
    #from modules.input import return_acceptable_states
    requirements = [s[0] for s in states]
    print(requirements)
    for k in requirements_dct:
        if requirements_dct[k] == requirements:
            
            return matrix_dct[k][0], matrix_dct[k][1], matrix_dct[k][2]
    #matrix3=matrix_dct[k][2]
    #print(states)
    #if requirements[0]=='E' or requirements[1]=='E':
    #    ssub=[]
    #    acceptable=return_acceptable_states(symmetry)
    #    for k in range(len(acceptable.full)):
    #            if requirements[0]=='E' and requirements[1]!='E':
    #                    if states[0]==acceptable.full[k]:
    #                            ssub.append(acceptable.labels[k][1])
    #                            break
    #            if requirements[0]!='E' and requirements[1]=='E':
    #                    if states[1]==acceptable.full[k]:
    #                            ssub.append(acceptable.labels[k][1])
    #                            break
    #            if requirements[0]=='E' and requirements[1]=='E':
    #                    if states[0]==acceptable.full[k]:
    #                            ssub.append(acceptable.labels[k][1])
    #                    if states[1]==acceptable.full[k]:
    #                            ssub.append(acceptable.labels[k][1])
        
    #    if len(ssub)==1:


    
    print('Error: Could not find matrix form for states '+str(states)+'.')
    exit()




import vhegen.modules.glbls as glo
import vhegen.modules.input as inp
import vhegen.modules.electronic as el
import vhegen.modules.vibrational as vi
import vhegen.modules.output as out
from copy import deepcopy
from time import time

class VHEGEN:
    def __init__(self,problem_dct):
        #initializes VHEGEN input parameters
        self.symmetry = problem_dct['sym']
        self.states = problem_dct['states'] 
        self.modes = problem_dct['modes']
        self.orders = problem_dct['o']
        self.filename = problem_dct['f']
        self.e_coords = 'both' #default e coordinate system
        self.basis = 'both' #default E state basis

    def set_basis(self,basis):
        if 'E' in [i[0] for i in self.states]:
            allowed_bases = ['complex','real','both']
            sat = False
            for b in allowed_bases:
                if b in str(basis):
                    self.basis = basis
                    sat = True
            if sat == False:
                raise Exception('EBasisError: E basis input must be "real", "complex", or "both".')

    def set_e_coordinates(self,coord_system):
        if hasattr(self,"expansions"):
            raise Exception('ECoordError: e mode coordinate system must be specified before expansion.')
        else:
            if 'E' in [i[0] for i in self.modes]:
                allowed_coords = ['pol','cart','both']
                sat = False
                for c in allowed_coords:
                    if c in str(coord_system):
                        self.e_coords = c
                        sat = True
                if sat == False:
                    raise Exception('ECoordError: e mode coordinate system must be "pol", "cart", or "both".')

    def return_init(self):
        #returns formatted input; for logging
        return inp.return_problem(self.symmetry,self.states,self.modes,self.orders,self.filename)

    def return_eigenvals(self):
        #returns formatted eigenvals; for logging
        string='Independent matrix elements: Eigenvals\n\n'
        for e in self.eigenvals:
            string += ('H_'+str(e)+': '+str(self.eigenvals[e]))+'\n'
        return string

    def get_constraints(self):
        self.constraints = []#vi.load_constraints(self.symmetry,self.eigenvals,self.modes)

    def get_formulas(self):
        self.__formulas = {}
        self.formulas = {}
        mapcoef={}
        k=0
        a=97
        l=len(self.matrix.original_mat)
        for i in range(1,5):
            if (i*i==l):
                l=i
                break
        for c,e in enumerate(self.eigenvals):
            v=0
            for c1,e1 in enumerate(self.matrix.original_mat):
                if (e1==e):
                    v=c1+1
                    col=(v-1)//l
                    break
            mapcoef[e]=chr(a+v-1-sum([k for k in range(col+1)]))
        for e in self.eigenvals:
            formula = vi.get_root_formula(self.eigenvals[e],self.modes,self.symmetry)
            #print(mapcoef[e])
            for i in range(len(formula)):   
                formula[i].coeff.letter=mapcoef[e]
            #New formula generator does not care about number of modes
            #if len(self.modes) < 2:
            #    formula = vi.adapt_to_unimodal(formula,self.modes)
            self.__formulas[e] = formula
            self.formulas[e] = vi.get_symbolic_formula(formula)
        #self.get_constraints()
        self.m_e_info()

    def return_formulas(self):
        string = ('Root formulas:\n\n')
        for e in self.formulas:
            string += ('H_'+str(e)+' : ' + str(self.formulas[e])+'\n')
        return string

    def return_constraints(self):
        return out.format_constraints(self.constraints)

    def init_matrix_form(self):
        self.matrix = el.VibronicMatrix(self.symmetry,self.states)
        self.eigenvals = self.matrix.symmetry_eigenvals
        self.matrix.get_dependencies()

    def get_expansions(self):
        expansions = {}
        count = {}
        parameters = {}
        element_position={}
        


        for o in self.orders:
            print('Order: '+str(o))
            expansions[o] = {}
            count[o] = {}
            parameters[o] = {}
            for e in self.eigenvals: #generate independent matrix elements
                print('Expanding H_{'+str(e).replace('__',',')+'}')
                #for e_coords='both', get_matrix_element_expansions returns normal structure but list of expansions for each matrix element, 1st being pol, 2nd cart.
                expansions[o][e],parameters[o][e] = vi.get_matrix_element_expansion(self.__formulas[e],o,self.e_coords,self.eigenvals[e],self.symmetry,self.modes)
                count[o][e] = len(parameters[o][e])
                #print([expansions[o][e][c] for c,i in enumerate(expansions[o][e])])
        expansions = el.get_dependent_elements(expansions,count,self) #map dependent matrix elements

        self.expansions = expansions
        self.parameters = parameters
        self.count = count
        if self.basis == 'real' or self.basis == 'both':
            out.log_append(self.basis_transformation())

    def return_termcount(self):
        if hasattr(self,'count'):
            count_string = ''
            if self.basis == 'complex' or self.basis == 'both':
                count_string += '\nNumber of fitting parameters (complex): \n\n'
                for o in self.count:
                    count_string += 'Order '+str(o)+':\n'
                    for e in self.count[o]:
                        count_string += 'H_'+str(e) +': '+str(self.count[o][e]) +'\n'
            if self.basis == 'real' or self.basis == 'both' and 'E' in [s[0] for s in self.states]:
                count_string += '\nNumber of fitting parameters (real): \n\n'
                for o in self.real_count:
                    count_string += 'Order '+str(o)+':\n'
                    for e in self.real_count[o]:
                        count_string += 'H_'+str(e) +': '+str(self.real_count[o][e]) +'\n'
            return count_string
        else:
            raise Exception('TermCountError: Can only count terms after expansion.')
        
    def return_expansions(self):
        two_coords = False
        if len(list(list(self.expansions.values())[0].values())[0]) > 1:
            two_coords = True
        string = 'Matrix element expansions: \n\n'
        for o in self.orders:
            string += 'Order: '+str(o)+'\n\n'
            if two_coords == True:
                string += ('Polar coordinates:\n\n')
            for e in self.expansions[o]:                    
                string += 'H_'+str(e)+' = '+str(self.expansions[o][e][0])+'\n\n'
            if two_coords == True:
                string += ('Cartesian coordinates:\n\n')
                for e in self.expansions[o]:
                    string += 'H_'+str(e)+' = '+str(self.expansions[o][e][1])+'\n\n'
        return string

    def basis_transformation(self):
        if 'E' in [i[0] for i in self.states]:
            if self.basis == 'real' or self.basis =='both':
                if self.basis =='real':
                    mapping = self.matrix.change_basis()
                else:
                    self.matrix_real = deepcopy(self.matrix)
                    mapping = self.matrix_real.change_basis()
                expansions_real,real_count = el.map_elements(self.expansions,mapping)
                self.real_count = real_count
            else:
                return ''
            if self.basis == 'real':
                self.expansions = expansions_real
            elif self.basis == 'both':
                self.expansions_real = expansions_real
            map_string = 'Complex to real basis transformation:\n\n'
            for e in mapping:
                map_string += str(e) +' = ' + str(mapping[e]) +'\n\n'
            return map_string
        else:
            self.matrix.returnpos()

    def auto(self):
        self.init_matrix_form()
        self.get_formulas()
        self.get_expansions()
        self.basis_transformation()

    def m_e_info(self):
        me_info = {}
        for e in self.eigenvals:
            #print(e)
            me_info[e] = [self.eigenvals[e],self.formulas[e]]
        self.matrix_element_info = me_info

    def convert(self,sympy_expr,syntax): #valid syntaxs include: 'LaTeX'
        converted_expansions = out.convert_syntax(sympy_expr,syntax)
        return converted_expansions

    def mctdhoutput(self,path='outputs'):
        if 'E' not in [s[0] for s in self.states]: #expansions is real already
            for o in self.orders:
                print(f'creating mctdh element for order {o}')
                if self.e_coords == 'both':
                    for e in self.expansions[o]:
                        if (self.expansions[o][e][0]!=0):
                            out.convert_to_MCTDH(self.expansions[o][e][0].expand(trig=True),e,self.matrix.element_position,self,'pol')
                        else:
                            out.convert_to_MCTDH(self.expansions[o][e][0],e,self.matrix.element_position,self,'pol')
                    for e in self.expansions[o]:
                        if (self.expansions[o][e][1]!=0):
                            out.convert_to_MCTDH(self.expansions[o][e][1].expand(),e,self.matrix.element_position,self,'cart')
                        else:
                            out.convert_to_MCTDH(self.expansions[o][e][1],e,self.matrix.element_position,self,'cart')        
                else:
                    for e in self.expansions[o]:
                        if (self.expansions[o][e][0]!=0):
                            out.convert_to_MCTDH(self.expansions[o][e][0].expand(trig=True),e,self.matrix.element_position,self,self.e_coords)
                        else:
                            out.convert_to_MCTDH(self.expansions[o][e][0],e,self.matrix.element_position,self,self.e_coords)         

        if self.basis == 'both' and 'E' in [s[0] for s in self.states]: #use expansions_real
            for o in self.orders:
                print(f'creating mctdh element for order {o}')
                if self.e_coords == 'both':
                    for e in self.expansions_real[o]:
                        if (self.expansions_real[o][e][0]!=0):
                            out.convert_to_MCTDH(self.expansions_real[o][e][0].expand(trig=True),e,self.matrix_real.element_position,self,'pol')
                        else:
                            out.convert_to_MCTDH(self.expansions_real[o][e][0],e,self.matrix_real.element_position,self,'pol')
                    for e in self.expansions_real[o]:
                        if (self.expansions_real[o][e][1]!=0):
                            out.convert_to_MCTDH(self.expansions_real[o][e][1].expand(),e,self.matrix_real.element_position,self,'cart')
                        else:
                            out.convert_to_MCTDH(self.expansions_real[o][e][1],e,self.matrix_real.element_position,self,'cart')
                else:
                    for e in self.expansions_real[o]:
                        if (self.expansions_real[o][e][0]!=0):
                            out.convert_to_MCTDH(self.expansions_real[o][e][0].expand(trig=True),e,self.matrix_real.element_position,self,self.e_coords)
                        else:
                            out.convert_to_MCTDH(self.expansions_real[o][e][0],e,self.matrix_real.element_position,self,self.e_coords)
        
            
        self.mctdhoutp.finalize()
        out.mctdh_write(self.filename+'-pol',path,self.mctdhoutp.fileform,self.mctdhoutp.inpfile)
        #self.mctdhoutp.write_to_screen()
        self.mctdhout.finalize()
        out.mctdh_write(self.filename+'-cart',path,self.mctdhout.fileform,self.mctdhout.inpfile)
        #self.mctdhout.write_to_screen()

    def pdflatex(self,path='outputs'):
        TeX_expansions = ''

        for o in self.orders:
            TeX_expansions += R'\subsection{Order: '+str(o)+R'}'+'\n'
            if self.basis == 'real':
                TeX_expansions += R'Number of fitting parameters: ' + out.realcount_format2TeX(self.real_count[o]) + '\n'
            else:
                TeX_expansions += R'Number of fitting parameters: ' + out.count_format2TeX(self.count[o]) + '\n'
            if self.e_coords == 'both':
                TeX_expansions += R'\subsubsection*{Polar e-coordinates:}'+'\n'
                for e in self.expansions[o]:
                    TeX_expansions += '\n'+R'\begin{align*}'+'\n'+self.convert(el.format_matrix_element(e),'LaTeX')+'^{('+str(o)+')}='+'&'+self.convert(self.expansions[o][e][0],'LaTeX')+'\n'+R'\end{align*}' +'\n'
                TeX_expansions += R'\subsubsection*{Cartesian e-coordinates:}' + '\n'
                for e in self.expansions[o]:
                    TeX_expansions += '\n'+R'\begin{align*}'+'\n'+self.convert(el.format_matrix_element(e),'LaTeX')+'^{('+str(o)+')}='+'&'+self.convert(self.expansions[o][e][1],'LaTeX')+'\n'+R'\end{align*}' +'\n'
            else:
                for e in self.expansions[o]:
                    TeX_expansions += '\n'+R'\begin{align*}'+'\n'+self.convert(el.format_matrix_element(e),'LaTeX')+'^{('+str(o)+')}='+'&'+self.convert(self.expansions[o][e][0],'LaTeX')+'\n'+R'\end{align*}' +'\n'
                    
        
        TeX_problem = out.problem_format2TeX(self.states,self.modes)

        if self.basis == 'both' and 'E' in [s[0] for s in self.states]: #Create two subsections
            TeX_expansions_real = ''
            for o in self.orders:
                TeX_expansions_real += R'\subsection{Order: '+str(o)+R'}'+'\n'
                TeX_expansions_real += R'Number of fitting parameters: ' + out.realcount_format2TeX(self.real_count[o]) + '\n'
                if self.e_coords == 'both':
                    TeX_expansions_real += R'\subsubsection*{Polar e-coordinates:}'+'\n'
                    for e in self.expansions_real[o]:
                        TeX_expansions_real += '\n'+R'\begin{align*}'+'\n'+self.convert(el.format_matrix_element(e),'LaTeX')+'^{('+str(o)+')}='+'&'+self.convert(self.expansions_real[o][e][0],'LaTeX')+'\n'+R'\end{align*}' +'\n'
                    TeX_expansions_real += R'\subsubsection*{Cartesian e-coordinates:}' + '\n'
                    for e in self.expansions_real[o]:
                        TeX_expansions_real += '\n'+R'\begin{align*}'+'\n'+self.convert(el.format_matrix_element(e),'LaTeX')+'^{('+str(o)+')}='+'&'+self.convert(self.expansions_real[o][e][1],'LaTeX')+'\n'+R'\end{align*}' +'\n'
                else:
                    for e in self.expansions_real[o]:
                        TeX_expansions_real += '\n'+R'\begin{align*}'+'\n'+self.convert(el.format_matrix_element(e),'LaTeX')+'^{('+str(o)+')}='+'&'+self.convert(self.expansions_real[o][e][0],'LaTeX')+'\n'+R'\end{align*}' +'\n'
                        
            TeX_content = out.compose_TeX_both_bases(self.symmetry,self.modes,self.matrix.format_TeX_mat(),self.matrix_real.format_TeX_mat(),TeX_problem,TeX_expansions,TeX_expansions_real)
        else:
            TeX_content = out.compose_TeX(self.symmetry,self.modes,self.matrix.format_TeX_mat(),TeX_problem,TeX_expansions)
            
        out.TeX_write(self.filename,path,TeX_content)
        out.exec_pdflatex(self.filename,path)

#EOF



from argparse import ArgumentParser
import vhegen.modules.glbls as glo
import os.path

try:
    input = raw_input
except NameError:
    pass

def configure_parser():
    parser = ArgumentParser()
    parser.add_argument("--c",dest="c",
                        help="config file path, default = config.cfg",
                        metavar="CONFIG", default="config.cfg")
    parser.add_argument("--sym",dest="sym",
                        help="point group static input",
                        metavar="SYMMETRY")
    parser.add_argument("--states",dest="states",
                        help="electronic states static input",
                        metavar="STATES")
    parser.add_argument("--modes",dest="modes",
                        help="vibrational modes static input.",
                        metavar="MODES")
    parser.add_argument("--o",dest="o",
                        help="order(s) of expansion static input.",
                        metavar="ORDERS")
    parser.add_argument("--f",dest="f",
                        help="output filename. default = 'output'",
                        metavar="FILENAME", default="output")
    return parser

def read_config(path):
    if (os.path.exists(path)):
        configfile = open(path,"r").readlines()
        config = {}
        for line in configfile:
            line = line.replace(u'\n','')
            if line[0] == '#':
                pass
            else:
                arg = line.split('=')
                config[arg[0]] = arg[1]
        return config
    else:
        return {'input':'dynamic',
            'pdf_out':'true',
            'log_out':'true',
            'e_coords':'both',
            'basis':'both',
            'mctdh_out':'true'}

def symmetry_arg(arg):
    
    arg= arg.replace(" ","")
    arg = arg.upper()

    if arg == 'LIST':
        print('\nAvailable point groups:')
        for s in glo.all_groups:
            print(s)
        raise Exception()
    elif arg == 'EXIT':
        exit()
    #elif symmetry in glo.all_groups:
    #    return symmetry
    else:
        l=len(arg)
        if arg[0].isalpha():
            valid=('C','D','S')
            lettornum=[]
            lettornum.append('l')
            if arg[0] in valid:
                letter=arg[0]
            else:
                raise Exception('InputError: First letter must be in:'+'C'+'D'+'S')
        else:
                raise Exception('InputError: First input must be a letter')
        if (arg[1:4]=='INF'): 
            rot=0
            if (arg[0]=='C' and arg[4]=='V'):
                extra='V'
                refl=True
                return {'letter':letter,
                    'rot':rot,
                    'extra':extra,
                    'refl':refl,
                    'print':letter+'inf'+extra.lower()}
            elif (arg[0]=='D' and arg[4]=='H'):
                extra='H'
                refl=True
                return {'letter':letter,
                    'rot':rot,
                    'extra':extra,
                    'refl':refl,
                    'print':letter+'inf'+extra.lower()}
            else:
                raise Exception('Not a valid symmetry with inf rotation')
        for i in range(1,l):
            if arg[i].isdigit():
                lettornum.append('n')
            elif arg[i].isalpha():
                lettornum.append('l')
            else:
                raise Exception('InputError: Bad character')

        if lettornum[1]!='n':
            raise Exception('InputError: second character must be a number')
        if l==2:
            rot=int(arg[1])
            extra=''
        elif l==3:
            if lettornum[2]=='n':
                rot=int(arg[1:3])
                extra=''
            elif lettornum[2]=='l':
                rot=int(arg[1])
                extra=arg[2]
        elif l==4:
            if lettornum[2]=='n':
                rot=int(arg[1:3])
                if (lettornum[3]=='l'):
                    extra=arg[3]
                else:
                    rot=int(arg[1:4])
                    extra=''
            else:
                raise Exception('InputError: not a symmetry group')
        elif l==5:
            if lettornum[2]=='n' and lettornum[3]=='n':
                rot=int(arg[1:4])
                if (lettornum[4]=='l'):
                    extra=arg[4]
                else:
                    raise Exception('InputError: you want rotation symmetry over 999?')
            else:
                raise Exception('InputError: not a symmetry group') 
        else:
            raise Exception('InputError: you want rotation symmetry over 999?')       
        
        valid=('','H','V','D')
        if extra in valid:
            if extra=='D' and letter!='D':
                raise Exception('InputError: not a valid symmetry group')
            if extra=='V' and letter!='C':
                raise Exception('InputError: not a valid symmetry group')
            if letter=='S' and extra!='':
                raise Exception('InputError: not a valid symmetry group') 
            if letter=='S' and rot%2!=0:
                raise Exception('InputError: not a valid symmetry group') 

        else:
            raise Exception('InputError: Last character must be in'+' h,v,d, ') 
        
        if extra =='D' and rot%2==0:
            rot=rot*2
        
        refl=False
        return {'letter':letter,
            'rot':rot,
            'extra':extra,
            'refl':refl,
            'print':letter+str(rot)+extra.lower()}

class state:
    def __init__(self):
        self.full=[]
        self.labels=[]
        self.size=0

    def getsize(self):
        self.size=len(self.print)

def return_acceptable_states(sym):
    states=state()
    
    if (sym['letter']=='C'):
        if (sym['rot']==0): #Cinfv
            states.full.append('A1')
            states.labels.append(('A',1,0,''))
            states.full.append('A2')
            states.labels.append(('A',2,0,''))
            maxemode=999
            for i in range(maxemode):
                states.full.append('E'+str(i+1))
                states.labels.append(('E',i+1,0,''))
            sym['refl']=True
        elif (sym['rot']%2==0):
            if (sym['extra']==''):
                states.full.append('A')
                states.labels.append(('A',0,0,''))
                states.full.append('B')
                states.labels.append(('B',0,0,''))
                emax=sym['rot']//2-1
                for i in range(emax):
                    states.full.append('E'+str(i+1))
                    states.labels.append(('B',i+1,0,''))
            elif (sym['extra']=='V'):
                states.full.append('A1')
                states.labels.append(('A',1,0,''))
                states.full.append('B1')
                states.labels.append(('B',1,0,''))
                states.full.append('A2')
                states.labels.append(('A',2,0,''))
                states.full.append('B2')
                states.labels.append(('B',2,0,''))
                emax=sym['rot']//2-1
                for i in range(emax):
                    states.full.append('E'+str(i+1))
                    states.labels.append(('E',i+1,0,''))
                sym['refl']=True
                
            elif (sym['extra']=='H'):
                states.full.append('Ag')
                states.labels.append(('A',0,1,'g'))
                states.full.append('Bg')
                states.labels.append(('B',0,1,'g'))
                emax=sym['rot']//2-1
                for i in range(emax):
                    states.full.append('E'+str(i+1)+'g')
                    states.labels.append(('E',i+1,1,'g'))
                states.full.append('Au')
                states.labels.append(('A',0,2,'u'))
                states.full.append('Bu')
                states.labels.append(('B',0,2,'u'))
                for i in range(emax):
                    states.full.append('E'+str(i+1)+'u')
                    states.labels.append(('E',i+1,2,'u'))
        elif (sym['rot']%2==1):
            
            if (sym['extra']==''):
                #print(sym['print'])
                states.full.append('A')
                states.labels.append(('A',0,0,''))
                emax=(sym['rot']+1)//2-1
                for i in range(emax):
                    states.full.append('E'+str(i+1))
                    states.labels.append(('E',i+1,0,''))
            elif (sym['extra']=='V'):
                states.full.append('A1')
                states.labels.append(('A',1,0,''))
                states.full.append('A2')
                states.labels.append(('A',2,0,''))
                emax=(sym['rot']+1)//2-1
                for i in range(emax):
                    states.full.append('E'+str(i+1))
                    states.labels.append(('E',i+1,0,''))
                sym['refl']=True
            elif (sym['extra']=='H'):
                states.full.append("A'")
                states.labels.append(('A',0,1,"'"))
                emax=(sym['rot']+1)//2-1
                for i in range(emax):
                    states.full.append('E'+str(i+1)+"'")
                    states.labels.append(('E',i+1,1,"'"))
                states.full.append("A''")
                states.labels.append(('A',0,2,"''"))
                for i in range(emax):
                    states.full.append('E'+str(i+1)+"''")
                    states.labels.append(('E',i+1,2,"''"))
    elif (sym['letter']=='D'):
        if (sym['rot']==0):
            if (sym['extra']=='H'):
                states.full.append('A1g')
                states.labels.append(('A',1,1,'g'))
                states.full.append('A2g')
                states.labels.append(('A',2,1,'g'))
                emax=999
                for i in range(emax):
                    states.full.append('E'+str(i+1)+'g')
                    states.labels.append(('E',i+1,1,'g'))
                states.full.append('A1u')
                states.labels.append(('A',1,2,'u'))
                states.full.append('A2u')
                states.labels.append(('A',2,2,'u'))
                emax=999
                for i in range(emax):
                    states.full.append('E'+str(i+1)+'u')
                    states.labels.append(('E',i+1,2,'u'))
                sym['refl']=True

        elif (sym['rot']%2==0):
            if (sym['extra']=='H'):
                states.full.append('A1g')
                states.labels.append(('A',1,1,'g'))
                states.full.append('A2g')
                states.labels.append(('A',2,1,'g'))
                states.full.append('B1g')
                states.labels.append(('B',1,1,'g'))
                states.full.append('B2g')
                states.labels.append(('B',2,1,'g'))
                emax=sym['rot']//2-1
                for i in range(emax):
                    states.full.append('E'+str(i+1)+'g')
                    states.labels.append(('E',i+1,1,'g'))
                states.full.append('A1u')
                states.labels.append(('A',1,2,'u'))
                states.full.append('A2u')
                states.labels.append(('A',2,2,'u'))
                states.full.append('B1u')
                states.labels.append(('B',1,2,'u'))
                states.full.append('B2u')
                states.labels.append(('B',2,2,'u'))
                emax=sym['rot']//2-1
                for i in range(emax):
                    states.full.append('E'+str(i+1)+'u')
                    states.labels.append(('E',i+1,2,'u'))
                sym['refl']=True
                
            elif (sym['extra']=='D'):
                states.full.append('A1')
                states.labels.append(('A',1,0,''))
                states.full.append('A2')
                states.labels.append(('A',2,0,''))
                states.full.append('B1')
                states.labels.append(('B',1,0,''))
                states.full.append('B2')
                states.labels.append(('B',2,0,''))
                emax=sym['rot']-1
                for i in range(emax):
                    states.full.append('E'+str(i+1))
                    states.labels.append(('E',i+1,0,''))
                sym['refl']=True

        elif (sym['rot']%2==1):
            if (sym['extra']=='H'):
                states.full.append("A1'")
                states.labels.append(('A',1,1,"'"))
                states.full.append("A2'")
                states.labels.append(('A',2,1,"'"))
                emax=sym['rot']//2
                for i in range(emax):
                    states.full.append('E'+str(i+1)+"'")
                    states.labels.append(('E',i+1,1,"'"))
                states.full.append("A1''")
                states.labels.append(('A',1,2,"''"))
                states.full.append("A2''")
                states.labels.append(('A',2,2,"''"))
                emax=sym['rot']//2
                for i in range(emax):
                    states.full.append('E'+str(i+1)+"''")
                    states.labels.append(('E',i+1,2,"''"))
                sym['refl']=True
            elif (sym['extra']=='D'):
                states.full.append("A1g")
                states.labels.append(('A',1,1,'g'))
                states.full.append("A2g")
                states.labels.append(('A',2,1,'g'))
                emax=sym['rot']//2
                for i in range(emax):
                    states.full.append('E'+str(i+1)+'g')
                    states.labels.append(('E',i+1,1,'g'))
                states.full.append("A1u")
                states.labels.append(('A',1,2,'u'))
                states.full.append("A2u")
                states.labels.append(('A',2,2,'u'))
                emax=sym['rot']//2
                for i in range(emax):
                    states.full.append('E'+str(i+1)+'u')
                    states.labels.append(('E',i+1,2,'u'))
                sym['refl']=True
    elif (sym['letter']=='S'):
        if (sym['rot']%4==0):
            if (sym['extra']==''):
                states.full.append('A')
                states.labels.append(('A',0,0,''))
                states.full.append('B')
                states.labels.append(('B',0,0,''))
                emax=sym['rot']//2-1
                for i in range(emax):
                    states.full.append('E'+str(i+1))
                    states.labels.append(('E',i+1,0,''))
        elif (sym['rot']%4==2):
            if (sym['extra']==''):
                states.full.append('Ag')
                states.labels.append(('A',0,1,'g'))
                emax=sym['rot']//4
                for i in range(emax):
                    states.full.append('E'+str(i+1)+'g')
                    states.labels.append(('E',i+1,1,'g'))
                states.full.append('Au')
                states.labels.append(('A',0,2,'u'))
                for i in range(emax):
                    states.full.append('E'+str(i+1)+'u')
                    states.labels.append(('E',i+1,2,'u'))              
    return states

def printfancy(acceptable):
    l=len(acceptable)
    #print(l)
    j=0
    for i in range((l-1)//5):
        #print(acceptable[j:j+5])
        print(f'{j*5:3d}:{acceptable[j*5]:6} {j*5+1:3d}:{acceptable[j*5+1]:6} {j*5+2:3d}:{acceptable[j*5+2]:6} {j*5+3:3d}:{acceptable[j*5+3]:6} {j*5+4:3d}:{acceptable[j*5+4]:6}')
        j=j+1
    if (l-j*5==1):
        #print(acceptable[j*5::])
        print(f'{j*5:3d}:{acceptable[j*5]:6}')
    elif (l-j*5==2):
        print(f'{j*5:3d}:{acceptable[j*5]:6} {j*5+1:3d}:{acceptable[j*5+1]:6}')
    elif (l-j*5==3):
        print(f'{j*5:3d}:{acceptable[j*5]:6} {j*5+1:3d}:{acceptable[j*5+1]:6} {j*5+2:3d}:{acceptable[j*5+2]:6}')
    elif (l-j*5==4):
        print(f'{j*5:3d}:{acceptable[j*5]:6} {j*5+1:3d}:{acceptable[j*5+1]:6} {j*5+2:3d}:{acceptable[j*5+2]:6} {j*5+3:3d}:{acceptable[j*5+3]:6}')
    elif (l-j*5==5):
        print(f'{j*5:3d}:{acceptable[j*5]:6} {j*5+1:3d}:{acceptable[j*5+1]:6} {j*5+2:3d}:{acceptable[j*5+2]:6} {j*5+3:3d}:{acceptable[j*5+3]:6} {j*5+4:3d}:{acceptable[j*5+4]:6}')

def addones(arg,sym):
    #appends 1 to e states if input is 
    #Cnx, n<=4
    #Dnd, n<=3
    #Dnh, n<=4
    #Dn,  n<=4
    #Sn,  n<=4
    numarg=len(arg)
    if (sym['letter']=='C' and sym['rot']<=4):
        for i in range(numarg):
            if arg[i][0].upper()=='E':
                if (len(arg[i])>1):
                    if arg[i][1]!='1':
                        arg[i]=arg[i][0]+'1'+arg[i][1:]
                else:
                    arg[i]=arg[i]+'1'
    if (sym['letter']=='D' and sym['extra']=='' and sym['rot']<=4):
        for i in range(numarg):
            if arg[i][0].upper()=='E':
                if (len(arg[i])>1):
                    if arg[i][1]!='1':
                        arg[i]=arg[i][0]+'1'+arg[i][1:]
                else:
                    arg[i]=arg[i]+'1'
    if (sym['letter']=='D' and sym['extra']=='D' and sym['rot']<=3):
        for i in range(numarg):
            if arg[i][0].upper()=='E':
                if (len(arg[i])>1):
                    if arg[i][1]!='1':
                        arg[i]=arg[i][0]+'1'+arg[i][1:]
                else:
                    arg[i]=arg[i]+'1'
    if (sym['letter']=='D' and sym['extra']=='H' and sym['rot']<=4):
        for i in range(numarg):
            if arg[i][0].upper()=='E':
                if (len(arg[i])>1):
                    if arg[i][1]!='1':
                        arg[i]=arg[i][0]+'1'+arg[i][1:]
                else:
                    arg[i]=arg[i]+'1'
    if (sym['letter']=='S' and sym['rot']<=4):
        for i in range(numarg):
            if arg[i][0].upper()=='E':
                if (len(arg[i])>1):
                    if arg[i][1]!='1':
                        arg[i]=arg[i][0]+'1'+arg[i][1:]
                else:
                    arg[i]=arg[i]+'1'

    

def states_arg(arg,sym):
    arg = arg.upper()
    arg = arg.replace('"',"''")
    arg =  arg.replace(" ","")
    
    acceptable=return_acceptable_states(sym).full
    if arg == 'LIST':
        print('\nIrreps of '+sym['print']+':')
        #acceptable=return_acceptable_states(sym).full
        printfancy(acceptable)
        raise Exception()
    elif arg == 'EXIT':
        exit()
    if ',' in arg:
        states = arg.replace(" ","").split(',')
    elif '+' in arg:
        states = arg.replace(" ","").split('+')
    else:
        states = [arg.replace(" ","")]

    #print(states)
    addones(states,sym) 
    #print(states)
    #exit()
    max_cond = len(states)
    met_cond = 0
    for i in range(max_cond):
        if states[i].isdigit():
            if int(states[i])<len(acceptable):
                states[i]=acceptable[int(states[i])]
            else:
                raise Exception(f'InputError: Number for state {i+1} too large')
        
           
    #print(states)
    for i in range(len(states)):
        
        for a in acceptable:
            if (states[i].upper()==a.upper()):
                states[i]=a
                met_cond += 1
    if met_cond==max_cond:
        #print(states)
        return states
    if met_cond != max_cond:
        raise Exception('InputError: State(s) '+arg+' not valid irreps in '+str(sym["print"])+'.')

def modes_arg(arg,sym):
    arg = arg.upper()
    arg = arg.replace('"',"''")
    if arg == 'LIST':
        print('\nIrreps of '+sym+':')
        for e in glo.irrep_dct[sym]:
            print(e)
        raise Exception()
    elif arg == 'EXIT':
        exit()
    if ',' in arg:
        modes = arg.split(',')
    elif '+' in arg:
        modes = arg.split('+')
    else:
        modes = [arg]
    if len(modes) not in (1,2):
        raise Exception('InputError: Number of modes must be 1 or 2.')
    max_cond = len(modes)
    met_cond = 0
    for e in modes:
        if e in glo.irrep_dct[sym]:
            met_cond += 1
    if met_cond == max_cond:
        return reorder(modes)
    else:
        raise Exception('InputError: Modes(s) '+modes.lower()+' not valid irreps in '+symmetry+'.')

def orders_arg(arg):
    if arg.upper() == 'EXIT':
        exit()
    orders = arg.split(',')
    for i,order in enumerate(orders):
        try:
            orders[i] = int(order)
        except ValueError:
            raise Exception('InputError: Orders must be non-negative integers.')
    if len(orders) > 2:
        raise Exception('InputError: Invalid range of orders.')
    elif len(orders) == 2:
        if orders[1] <= orders[0]:
            raise Exception('InputError: Invalid range of orders.')
        if (orders[0]) < 0 or (orders[1] < 0):
            raise Exception('DynamicInputError: Orders must be non-negative integers.')
    elif len(orders) == 1:
        if orders[0] < 0:
            raise Exception('InputError: Orders must be non-negative integers.')
    if len(orders) == 2:
        orders = range(int(orders[0]),int(orders[1])+1)
    return orders
    
def filename_arg(arg):
    for char in (r"!@#$%^&*(){}[],."):
        if char in arg:
            print(char)
            raise Exception('InputError: Filename may only contain letters, numbers, "_", and "-".')
    if len(arg) == 0:
        return 'output'
    else:
        return arg

def format(inp_list):
    formatted_str = '('
    for i,n in enumerate(inp_list):
        if i>0 and i<len(inp_list):
            formatted_str+='+'
        formatted_str+=n
    formatted_str+=')'
    return formatted_str

def format_problem(states,modes):
    states_str = format(states)
    modes_str = format(modes)
    return states_str+'x'+modes_str.lower()

def reorder(inp_list):
    #re-orders states/modes by priority eg) (A+E)x(e+a) -> (E+A)x(e+a)
    if len(inp_list) > 1:
        if glo.irrep_priority.index(inp_list[0][0]) > glo.irrep_priority.index(inp_list[1][0]):
            inp_list = list(reversed(inp_list))
    return inp_list
    
def return_problem(symmetry,states,modes,orders,filename):
    return  ('\n---------------------------------------'
             '\n  VHEGEN instance parameters:'
             '\n  Symmetry: ' +symmetry['letter']+str(symmetry['rot'])+symmetry['extra']+
             '\n  Problem: ' + format_problem(states,modes)+
             '\n  Expansion orders: ' + str(orders)+
             '\n  Output filename: ' + filename+
             '\n---------------------------------------')

def dynamic_input():
    #symmetry
    dyn_inp = input('Enter symmetry: ')
    while True:
        try:
            sym_inp = symmetry_arg(dyn_inp)
        except Exception as e:
            print(e)
            dyn_inp = input('Re-enter symmetry: ')
            continue
        else:
            symmetry = sym_inp
            print(symmetry)
            print(symmetry['print']+' symmetry accepted.')
            break
    #states
    dyn_inp = input('Enter electronic state(s): ')
    while True:
        try:
            states_inp = states_arg(dyn_inp,symmetry)
            if len(states_inp) not in (1,2):
                raise Exception('InputError: Number of states must be 1 or 2.')
            #print(states_inp)
            states_inp = reorder(states_inp)
        except Exception as e:
            print(e)
            dyn_inp = input('Retry state(s): ')
            continue
        else:
            states = states_inp
            print('Electronic states '+ format(states)+' accepted.')
            break
    #modes
    dyn_inp = input('Enter vibrational mode(s): ')
    while True:
        try:
            modes_inp = states_arg(dyn_inp,symmetry)
            #modes_inp = reorder(modes_inp)
            #if symmetry in glo.trigonal_groups:
            #    modes_inp = modes_inp #limit_trigonal(modes_inp)
        except Exception as e:
            print(e)
            dyn_inp = input('Retry mode(s): ')
            continue
        else:
            modes = modes_inp
            print('Vibrational modes '+format(modes).lower()+' accepted.')
            break
    #orders
    dyn_inp = input('Order(s) of expansion:')
    while True:
        try:
            orders_inp = orders_arg(dyn_inp)
        except Exception as e:
            print(e)
            dyn_inp = input('Retry orders: ')
            continue
        else:
            orders = orders_inp
            print('Orders of expansion '+(str(list(orders)))+' accepted.')
            break
    #output filename
    dyn_inp = input('Enter filename: ')
    while True:
        try:
            filename_inp = filename_arg(dyn_inp)
        except Exception as e:
            print(e)
            dyn_inp = input('Retry filename: ')
            continue
        else:
            filename = filename_inp
            print('Filename: "'+filename+'".')
            break

    print(return_problem(symmetry,states,modes,orders,filename))

    while True:
        yn_prompt = input('Continue? [Y/N]: ')
        if yn_prompt.upper() == 'Y':
            print('')
            return symmetry, states, modes, orders, filename
        elif yn_prompt.upper() == 'N' or yn_prompt.upper() == 'EXIT':
            exit()

def read_input():
    argparser = configure_parser()
    options = argparser.parse_args()
    config = read_config('config.cfg')
    if 'static' in str(config[u'input']):
        print("Reading static inputs.")
        #static input
        try:
            symmetry = symmetry_arg(options.sym)
        except Exception as e:
            print('StaticInputError: Invalid symmetry argument.')
            exit()
        try:
            if len(options.states) not in (1,2):
                raise Exception('InputError: Number of states must be 1 or 2.')
            #states = reorder(states_arg(options.states,symmetry))
            #print(options.states)
            states = reorder(states_arg(options.states,symmetry))
            #states=reorder(states)
            #if len(states_inp) not in (1,2):
            #    raise Exception('InputError: Number of states must be 1 or 2.')
            #print(states_inp)
            #states_inp = reorder(states_inp)
            #if symmetry in glo.trigonal_groups:
            #    states = states #limit_trigonal(states)
        except Exception as e:
            print('StaticInputError: Invalid states argument.')
            exit()
        try:
            modes = states_arg(options.modes,symmetry)
            #modes = reorder(modes_arg(options.modes,symmetry))
            #if symmetry in glo.trigonal_groups:
            #    modes = modes #limit_trigonal(modes)
        except Exception as e:
            print(e)
            print('StaticInputError: Invalid modes argument.')
            exit()
        try:
            orders = orders_arg(options.o)
        except Exception as e:
            print('StaticInputError: Invalid orders argument.')
            exit()
        try:
            filename = filename_arg(options.f)
        except Exception as e:
            print('StaticInputError: Invalid filename argument.')
            exit()
    elif 'dynamic' in str(config[u'input']):
        print("Entering dynamic input.")
        #enter dynamic input
        symmetry,states,modes,orders,filename = dynamic_input()
    else:
        print("Error: 'input' value in config.cfg not found or recognized.")
        exit()
    problem_dct = {'sym':symmetry,
                   'states':states,
                   'modes':modes,
                   'o': orders,
                   'f': filename}
    return config,problem_dct

def prepare_input(sym,states,modes,orders,filename='output'):
    sym = symmetry_arg(sym)
    states = reorder(states_arg(states,sym))
    modes = states_arg(modes,sym)
    orders = orders_arg(orders)
    filename = filename_arg(filename)
    return {'sym':sym,
            'states':states,
            'modes':modes,
            'o': orders,
            'f':filename}
#EOF



from vhegen.modules.electronic import kdelta 
from vhegen.modules.input import return_acceptable_states
#, refl_parity, inver_parity
from sympy import sqrt, Symbol
from math import exp,pi

def return_eigenvals(symmetry, states):
	
	#print(states)
	acceptable=return_acceptable_states(symmetry)
	requirements = [s[0] for s in states]
	if len(states)>1:
		if requirements[0]=='E' or requirements[1]=='E':
			ssub=[]
			
			for k in range(len(acceptable.full)):
				if requirements[0]=='E' and requirements[1]!='E':
					if states[0]==acceptable.full[k]:
						ssub.append(acceptable.labels[k][1])
						break
				if requirements[0]!='E' and requirements[1]=='E':
					if states[1]==acceptable.full[k]:
						ssub.append(acceptable.labels[k][1])
						break
				if requirements[0]=='E' and requirements[1]=='E':
					if states[0]==acceptable.full[k]:
						ssub.append(acceptable.labels[k][1])
					if states[1]==acceptable.full[k]:
						ssub.append(acceptable.labels[k][1])
	else:
		if requirements[0]=='E':
			ssub=[]
			for k in range(len(acceptable.full)):
				if states[0]==acceptable.full[k]:
					ssub.append(acceptable.labels[k][1])
		

	n = symmetry['rot']
	labels=[]
	for i in range(len(states)):
		for k in range(len(acceptable.full)):
			if states[i]==acceptable.full[k]:
				labels.append(acceptable.labels[k])

	refl_parity=[]
	inver_parity=[]
	for i in range(len(states)):
		refl_parity.append(labels[i][1])

		if (labels[i][2]!=0):
			inver_parity.append(pow((-1),labels[i][2]+1))
		else:
			inver_parity.append(0)
	#print(refl_parity)
	#exit()
	if (len(states)==1):
		refl_parity.append(0)
		inver_parity.append(0)
		#print(labels)
		#print(refl_parity)
		#print(inver_parity)


	added_dummy = False
	if len(states) == 1:
		added_dummy = True
		states.append('dummystate')

	if n % 2 == 1: #(odd n-axial) problems

		requirements_dct = {0: ['A'],
							1: ['E'],
							2: ['A','A'],
							3: ['E','A'],
							4: ['E','E']}

		eigenvals_dct = {0: {Symbol('A__A'):[0,1,0,1]},

        				 1: {Symbol('+__+'): [0,1,0,1],
                             Symbol('+__-'): [2*refl_parity[0],1,-1,1]},

						 2: {Symbol('A_alpha__A_beta'):[0,(-1)**(kdelta(refl_parity[0],refl_parity[1])+1),0,(-1)**(kdelta(inver_parity[0],inver_parity[1])+1)]},
						   
						 3: {Symbol('+__A'):[refl_parity[0],(-1)**kdelta(refl_parity[1],2),(-1)**kdelta(refl_parity[1],1),(-1)**(kdelta(inver_parity[0],inver_parity[1])+1)]},
						   
						 4: {Symbol('+_alpha__+_beta'):[refl_parity[0]-refl_parity[1],1,-1,(-1)**(kdelta(inver_parity[0],inver_parity[1])+1)],
                             Symbol('+_alpha__-_beta'):[refl_parity[0]+refl_parity[1],1,-1,(-1)**(kdelta(inver_parity[0],inver_parity[1])+1)]}
        				   
                        }

	else: #(even n-axial) problems

		requirements_dct = {0: ['A'],
							1: ['B'],
							2: ['E'],
							3: ['A','A'],
							4: ['A','B'],
							5: ['E','A'],
							6: ['E','B'],
							7: ['E','E'],
							8: ['B','B']}

		eigenvals_dct = {0: {Symbol('A__A'):[0,1,0,1]},

						 1: {Symbol('B__B'):[0,1,0,1]},

						 2: {Symbol('+__+'): [0,1,0,1],
                             Symbol('+__-'): [refl_parity[0]*2,1,-1,1]},

						 3: {Symbol('A_alpha__A_beta'):[0,(-1)**(kdelta(refl_parity[0],refl_parity[1])+1),0,(-1)**(kdelta(inver_parity[0],inver_parity[1])+1)]},
							
						 4: {Symbol('A__B'):[n//2,(-1)**(kdelta(refl_parity[0],refl_parity[1])+1),0,(-1)**(kdelta(inver_parity[0],inver_parity[1])+1)]},

						 5: {Symbol('+__A'):[refl_parity[0],(-1)**kdelta(refl_parity[1],2),(-1)**kdelta(refl_parity[1],1),(-1)**(kdelta(inver_parity[0],inver_parity[1])+1)]},
						
						 6: {Symbol('+__B'):[refl_parity[0]+n//2,(-1)**kdelta(refl_parity[1],2),(-1)**kdelta(refl_parity[1],1),(-1)**(kdelta(inver_parity[0],inver_parity[1])+1)]},

						 7: {Symbol('+_alpha__+_beta'):[refl_parity[0]-refl_parity[1],1,-1,(-1)**(kdelta(inver_parity[0],inver_parity[1])+1)],
                             Symbol('+_alpha__-_beta'):[refl_parity[0]+refl_parity[1],1,-1,(-1)**(kdelta(inver_parity[0],inver_parity[1])+1)]}, 

						 8: {Symbol('B_alpha__B_beta'):[0,(-1)**(kdelta(refl_parity[0],refl_parity[1])+1),0,(-1)**(kdelta(inver_parity[0],inver_parity[1])+1)]}
						}
						
	if added_dummy == True:
		del states[-1]
		
	problem_reqs = [s[0] for s in states]

	for k in requirements_dct:
		if requirements_dct[k] == problem_reqs:
			return eigenvals_dct[k]

	print('Error: Could not find eigenvalues for states '+str(states)+'.')
	exit()

#EOF



import sympy as sym
import itertools
import os
from vhegen.modules.glbls import replace_all, escapes
from vhegen.modules.input import reorder
from vhegen.modules.input import return_acceptable_states
from copy import copy,deepcopy
import numpy as np

class Term: #ExpansionTerm
    def __init__(self,prefactor,coeff,funcs):
        self.prefactor = prefactor      #prefactor should be any sympy-digestible argument
        self.coeff = coeff              #coeff should be an instance of Coeff
        self.indices = coeff.indices
        self.funcs = funcs              #funcs should be a list of instances of Mono and/or Trig 
        self.form = [self.prefactor,self.coeff] + self.funcs
        l=len(funcs)
        self.sinorcos='neither'
        if isinstance(funcs[l-1],Cos):
            self.sinorcos='cos'
        if isinstance(funcs[l-1],Sin):
            self.sinorcos='sin'
    
    def change_to_exp(self):
        if self.sinorcos=='sin':
            self.prefactor=sym.I
            self.update_form()
            self.funcs[len(self.funcs)-1].change_to_exp()
        if (self.sinorcos=='cos'):
            self.funcs[len(self.funcs)-1].change_to_exp()
    #1  
    def update_indices(self,p):
        #need to update them in coeff and funcs
        for c,i in enumerate(self.indices): #count, index
            #print(self.indices[c])
            if (i[0]=='K'):
                self.coeff.indices[c] = self.coeff.indices[c].replace(i,str(p[c]//2))
            else:
                self.coeff.indices[c] = self.coeff.indices[c].replace(i,str(p[c]))
            #print(self.indices[c])
            for c2,f in enumerate(self.funcs):
                if isinstance(f,Mono):
                    if (i[0]=='K'):
                        f.arg = f.arg.replace(i,str(p[c]//2))
                    else:
                        f.arg = f.arg.replace(i,str(p[c]))
                elif isinstance(f,Trig):
                    for c3,a in enumerate(f.args):
                        f.args[c3] = a.replace(i,str(p[c]))

    def refresh_indices(self):
        self.indices = self.coeff.indices

    def update_form(self):
        self.form = [self.prefactor,self.coeff] + self.funcs

    def get_order(self):
        order = 0
        for f in self.funcs:
            if isinstance(f,Mono):
                order += sym.sympify(f.arg)
        self.order = order

    def zero(self):
        self = 0

    def grab_constraints(self,constraint_dct): #checks global constraints, acquires relevant("local") constraints, returns them in index-position basis.
        applicable_constraints = {}
        for c in constraint_dct:
            if 'nz' in constraint_dct[c]:
                applicable_constraints[c] = constraint_dct[c]
            if c == 'all':
                applicable_constraints[c] = constraint_dct[c]
            else: #summing-index based constraints
                if '&' in c:
                    c_parse = c.split('&')
                    c_index = []
                    for i in self.indices:
                        for t in c_parse:
                            if t in i:
                                c_index.append(self.indices.index(i))
                    if len(c_index) == 2:
                        applicable_constraints['&'.join([str(i) for i in c_index])] = constraint_dct[c]
                else:
                    for i in self.indices:
                        if c in i:
                            if self.indices.index(i) in applicable_constraints:
                                applicable_constraints[self.indices.index(i)].add(constraint_dct[c])
                            else:
                                applicable_constraints[self.indices.index(i)] = set([constraint_dct[c]])
        self.local_constraints = applicable_constraints
        return self.local_constraints

    #def constrained_product(val_list):


    #2  
    def expand(self,order,eigenvals,symmetry,modes):
        from modules.input import return_acceptable_states
        #order: order of expansion
        #eigenvals: list of eigenvalue numbers
        
        acceptable=return_acceptable_states(symmetry)
        modelabels=[]
        for k in range(len(modes)):
            for m in range(len(acceptable.full)):
                if acceptable.full[m]==modes[k]:
                    modelabels.append(acceptable.labels[m])
        
        indpos, indlabels = assign_indices(modes,modelabels)

        #print(self.prefactor)
        #print(self.coeff.stem)
        #print(indpos)
        #print(indlabels)
        #exit()
        #print(modes)
        #print(modelabels)
        #exit()

        val_list = gen_index_lists(self.indices,order,eigenvals,symmetry['rot'])
        partitions = []
        expansions = [] #index of the partitions matches index of the summing indices
        nm_constraint = False
        needrefl=False
        #print(symmetry)
        if symmetry['refl']:
            needrefl=True
        needexpand=True
        if needrefl and self.coeff.stem=='r' and self.sinorcos=='sin' and eigenvals[2]==0:
            needexpand=False
        if needrefl and self.coeff.stem=='i' and self.sinorcos=='cos' and eigenvals[2]==0:
            needexpand=False
        if needrefl and self.coeff.stem=='i' and self.sinorcos=='neither' and eigenvals[2]==0:
            needexpand=False
        
        #vals_iterprod = list(itertools.product(*val_list))
        if needexpand:
            for c,n in enumerate(itertools.product(*val_list)):
                meets_real_req = True
                nabs=[abs(n[i]) for i in range(len(n))]
                if (sum(nabs) == order) and (n not in partitions):
                    
                    meets_req = apply_constraints_new(n,self.prefactor,symmetry,indpos,indlabels,eigenvals,self.sinorcos)
                    partitions.append(n)              
                    #if meets_req == True:
                    #    #print(n)
                    #    #exit()
                    if meets_req == True and meets_real_req == True:
                        term = deepcopy(self)
                        term.update_indices(n)
                        term.get_order()
                        expansions.append(term)
        self.expansions = expansions

    def expand_new(self,order,eigenvals,symmetry,modes):
        
        #order: order of expansion
        #eigenvals: list of eigenvalue numbers
        
        acceptable=return_acceptable_states(symmetry)
        modelabels=[]
        for k in range(len(modes)):
            for m in range(len(acceptable.full)):
                if acceptable.full[m]==modes[k]:
                    modelabels.append(acceptable.labels[m])
        
        indpos, indlabels = assign_indices(modes,modelabels)
        mult, omult = assign_multipliers(modes,modelabels)
        #omult=gen_real_restraint(omult,self.indices,order,eigenvals,symmetry['rot'])
        #print(mult)
        #print(omult)

        #val_list = gen_index_lists(self.indices,order,eigenvals,symmetry['rot'])
        partitions = []
        expansions = [] #index of the partitions matches index of the summing indices
        nm_constraint = False
        needrefl=False
        #print(symmetry)
        realexpand=False
        needexpand=True
        if eigenvals[0]==0:
            realexpand=True
            if self.prefactor==1j:
                needexpand=False
        if symmetry['rot']>0:
            if eigenvals[0]*2==symmetry['rot']:
                realexpand=True
                if self.prefactor==1j:
                    needexpand=False
        if symmetry['refl']:
            needrefl=True

        if needrefl and self.coeff.stem=='r' and self.sinorcos=='sin' and eigenvals[2]==0:
            needexpand=False
        if needrefl and self.coeff.stem=='i' and self.sinorcos=='cos' and eigenvals[2]==0:
            needexpand=False
        if needrefl and self.coeff.stem=='i' and self.sinorcos=='neither' and eigenvals[2]==0:
            needexpand=False
        
        expansions=[]
        indices=[]
        #vals_iterprod = list(itertools.product(*val_list))
        if needexpand:
            nmodes=len(self.indices)
            str_builder=-1*np.ones(nmodes,dtype=int)
            prevd=-1
            orderin=copy(order)
            previ=-1*np.ones(nmodes,dtype=int)
            list_of_terms=[]
            self.expand_for_order(nmodes,str_builder,previ,prevd,orderin,mult,omult,symmetry,indpos,indlabels,eigenvals,list_of_terms)
            #exit()
            for i in range(len(list_of_terms)):
                goodterm=True
                if realexpand:
                    goodterm=gen_real_restraint(list_of_terms[i],self.indices,order,eigenvals,symmetry['rot'])
                if (goodterm):
                    term = deepcopy(self)
                    term.update_indices(list_of_terms[i])
                    term.get_order()
                    expansions.append(term) 
                    indices.append(term.coeff.indices)
                    #if meets_req == True and meets_real_req == True:
                    #    term = deepcopy(self)
                    #    term.update_indices(n)
                    #    term.get_order()
                    #    expansions.append(term)
        self.expansions = expansions
        self.indiceslist=indices

    def expand_for_order(self,Nd,str_builder,previ,prevd,order,mult,omult,symm,ip,il,evals,lot):
        if (prevd==Nd-1):
            if (order==0):
                #print(str_builder)
                #print(self.coeff.stem)
                meets_req = apply_constraints_new(str_builder,self.prefactor,symm,ip,il,evals,self.sinorcos)
                
                if (meets_req):
                    #print([self.prefactor,self.coeff.stem,self.sinorcos,str_builder*mult]) 
                    lot.append(np.ndarray.tolist(str_builder*mult))  
                return
            else:
                return
        else:
            newi=np.ones(Nd,dtype=int)
            d=prevd+1 #go up a dimension
            newi=-1*np.ones(Nd,dtype=int)
            newi[0:d+1]=previ[0:d+1]
            for i in range(previ[d]+1,order*omult[d]-order//2*(mult[d]-1)+1):
                if (omult[d]==2):
                    str_builder[d]=(i+1)//2*(-1)**i
                else:
                    str_builder[d]=i
                newi[d]=i
                self.expand_for_order(Nd,str_builder,newi,d,order-abs(str_builder[d]*mult[d]),mult,omult,symm,ip,il,evals,lot)
            d=d-1 #go back a dimension


    #3
    def compile_expansions(self):   #compile expansion term as sympy expr
        collect_cartesian_terms = None
        for func in self.funcs:
            if isinstance(func,Trig):
                trigfunc = func
                if trigfunc.cartesian == True:
                    if len(trigfunc.args) > 1:
                        collect_cartesian_terms = 2
                    else:
                        collect_cartesian_terms = 1
        free_parameters = []
        compiled_expansions = []
        for term in self.expansions:
            term.coeff.compute()
            term.coeff.compile()
            
            for i,func in enumerate(term.funcs):
                func.compute()
                func.compile()
            compiled_term = 1
            for part in term.form:
                if hasattr(part,'sympy_form') == False:
                    compiled_term = compiled_term*part
                else:
                    compiled_term = compiled_term*part.sympy_form
            #if two e modes:
            if collect_cartesian_terms == 2:
                symlist=[]
                symlist.append('x_1')
                symlist.append('y_1')
                for i in range(1,len(trigfunc.args)):
                    symlist.append(f'x_{i+1}')
                    symlist.append(f'y_{i+1}')
                #compiled_term = sym.collect(compiled_term,['x_alpha','y_alpha','x_beta','y_beta'],exact=True)
                compiled_term = sym.collect(compiled_term,symlist,exact=True)
            #else if 1 e mode:
            elif collect_cartesian_terms == 1:
                compiled_term = sym.collect(compiled_term,['x_1','y_1'],exact=True) 
            compiled_expansions.append(compiled_term)

            if term.coeff.sympy_form not in free_parameters and compiled_term != 0:
                free_parameters.append(term.coeff.sympy_form)
        #self.expansions_old=deepcopy(self.expansions)
        self.expansions = compiled_expansions
        self.parameters = free_parameters

    def compile_formula(self):
        deepcopyself = deepcopy(self)
        sym_formula = ''
        if deepcopyself.prefactor != 1:
            prefactor = str(deepcopyself.prefactor)
            prefactor = prefactor.replace('1j','i')
            sym_formula += ' '+prefactor+'*'
        sym_formula += str(deepcopyself.coeff.compile())
        for func in deepcopyself.funcs:
            func.symcomp()
            sym_formula += func.symb
        return sym_formula
    
    def convert_to_cartesian(self):
        for part in self.funcs:
            part.convert_to_cartesian()

    def convert_to_polar(self):
        for part in self.funcs:
            part.convert_to_polar()

    def adapt_to_unimodal(self,modes):

        if modes[0][0] != 'A': #gamma+a to gamma
            del self.coeff.indices[0]
            self.refresh_indices()
            i = 0
            max_i = len(self.funcs) - 1

            while i <= max_i:
                if isinstance(self.funcs[i],Mono):
                    if self.funcs[i].coord == 'z':
                        del self.funcs[i]
                        i -= 1
                        max_i -= 1

                i += 1

        else: #a+a to a
            for c,i in enumerate(self.indices):
                if 'b' in i:
                    del self.coeff.indices[c]
            self.refresh_indices()

            i = 0
            max_i = len(self.funcs) - 1

            while i <= max_i:
                if isinstance(self.funcs[i],Mono):
                    if self.funcs[i].label == '_beta':
                        del self.funcs[i]
                        i -= 1
                        max_i -= 1
                    else:
                        self.funcs[i].label = ''
                i += 1

        self.update_form()

def assign_indices(modes,modelabels):
    #label and count each modes
    posa=[]
    posb=[]
    pose=[]
    modes0=[mode[0] for mode in modes]
    for i in range(len(modes)):
        if modes0[i]=='E':
            pose.append(i)
        if modes0[i]=='A':
            posa.append(i)
        if modes0[i]=='B':
            posb.append(i)
    
    indlabels=[]
    indpos=[]
    j=0
    for i in range(len(posa)):
        indlabels.append(modelabels[posa[i]])
        indpos.append(j)
        j=j+1
    for i in range(len(posb)):
        indlabels.append(modelabels[posb[i]])
        indpos.append(j)
        j=j+1
    for i in range(len(pose)):
        indlabels.append(modelabels[pose[i]])
        indpos.append(j)
        j=j+2 #skip K indices because they don't contribute to constraints outside of order
    return indpos, indlabels

def assign_multipliers(modes,modelabels):
    #label and count each modes
    posa=[]
    posb=[]
    pose=[]
    modes0=[mode[0] for mode in modes]
    tmodes=0
    for i in range(len(modes)):
        if modes0[i]=='E':
            pose.append(i)
            tmodes=tmodes+2
        if modes0[i]=='A':
            posa.append(i)
            tmodes=tmodes+1
        if modes0[i]=='B':
            posb.append(i)
            tmodes=tmodes+1
    

    mult=np.ones((tmodes),dtype=int)
    omult=np.ones((tmodes),dtype=int)
    j=0
    for i in range(len(posa)):
        mult[j]=1
        omult[j]=1
        j=j+1
    for i in range(len(posb)):
        mult[j]=1
        omult[j]=1
        j=j+1
    for i in range(len(pose)):
        mult[j]=1 #M index
        omult[j]=2 #
        mult[j+1]=2 #K index
        omult[j+1]=1
        j=j+2 #skip K indices because they don't contribute to constraints outside of order
    return mult, omult

def rdelta(p,q):
    if p==q:
        return 1
    else:
        return 0

def apply_constraints_new(n,prefactor,symmetry,indpos,indlabels,eigenvals,sinorcos):
    #n:  list of indices
    #prefactor: either +1,-1 or 1j
    #symmetry: dictionary of symmetry properties
    #indpos: list of index positions that matter, i.e. skipping K
    #indlabels: list of labels for each indices with symmetry properties
    #eigenvals: the list of four eigenvalues that define the system
    #sinorcos: states whether expansion function has sin or cos or neither
    kappa=eigenvals[0]
    rot=symmetry['rot']
    s=kappa
    #print(n)
    for i in range(len(indlabels)):
        if indlabels[i][0]=='B':
            s=s+n[indpos[i]]*(rot//2)
        if indlabels[i][0]=='E':
            s=s+n[indpos[i]]*indlabels[i][1]
        #print('rot')
        #print((s,s//rot))
    if (rot!=0):
        if (s//rot)*rot!=s:
            #print('rot not followed')
            return False
    else:
        if (s!=0):
            return False
        

    #print(isinstance(prefactor, complex))
    needrefl=False
    if symmetry['refl']:
        needrefl=True
        if isinstance(prefactor, complex):
            refleig=eigenvals[2]
        else:
            refleig=eigenvals[1]
        #print(refleig)
        if refleig==1:
            refl_start=0
        else: #refleig=-1
            refl_start=1

    
    if needrefl: #is reflection symmetry needed

        if sinorcos=='sin': #phi->-phi flips sign
            refl_start=refl_start+1
        else: #phi->-phi does nothing
            refl_start=refl_start+0
        s=refl_start
        for i in range(len(indlabels)):
            if indlabels[i][0]=='A' or indlabels[i][0]=='B':
                s=s+n[indpos[i]]*rdelta(2,indlabels[i][1])
        #print('refl')
        #print((s,s//2))
        if (s//2)*2!=s:
            #print('reflect not followed')
            #print(sinorcos)
            #print(prefactor)
            return False
    
    if indlabels[0][2]!=0: #inver symmetry needed
        if eigenvals[3]==1:
            inver_start=0
        else:
            inver_start=1
        s=inver_start
        for i in range(len(indlabels)):
            s=s+n[indpos[i]]*rdelta(2,indlabels[i][2])
        #print('inver')
        #print(n)
        #print(inver_start)
        #print((s,s//2))
        if (s//2)*2!=s:
            return False
    return True

        

class Coeff:
    def __init__(self,letterstem,indices):
        self.letter = letterstem[0]
        self.stem = letterstem[1]
        self.indices = indices
        self.indroot = [indi[0:2] for indi in indices]

    def compute(self):
        for i,index_arg in enumerate(self.indices):
            self.indices[i] = sym.sympify(index_arg)

    def compile(self):
        #compile into sympy symbol
        indexstr = '_{'
        for c,s in enumerate(self.indices):
            indexstr+=str(s)
            if c != (len(self.indices) -1):
                indexstr+=','
            else:
                indexstr += '}'
        self.sympy_form = sym.Symbol(self.letter+'^{'+self.stem+'}'+indexstr,real=True)
        return self.sympy_form

    #def sympy_coeff(self):
    #    indexstr = '_{'
    #   for c,s in enumerate(self.indices):
    #        indexstr+=str(s)
    #        if c != (len(self.indices) -1):
    #            indexstr+=','
    #        else:
    #            indexstr += '}'
    #    return sym.Symbol(self.letter+'^{'+self.stem+'}'+indexstr,real=True)

class Mono:
    def __init__(self,coord,label,arg):
        self.coord = coord
        self.label = label
        self.arg = arg
        self.cartesian = False

    def compute(self):
        self.arg = sym.simplify(self.arg)

    def convert_to_cartesian(self):
        if self.coord == 'rho':
            self.coord = 'x,y'
            self.cartesian = True

    def convert_to_polar(self):
        if self.coord == 'x,y':
            self.coord = 'rho'
            self.cartesian = False

    def symcomp(self):
        self.symb = '*'+self.coord+self.label + '**('+ self.arg +')'

    def compile(self):
        if self.cartesian != True:
            self.sympy_form = sym.Symbol(self.coord+self.label,real=True)**(self.arg)
        else:
            expr = (sym.sqrt(sym.Symbol('x'+self.label,real=True)**2 + sym.Symbol('y'+self.label,real=True)**2))**self.arg #(sqrt(x**2 + y**2))
            self.sympy_form = sym.factor(sym.expand(expr)) #sym.factor

class Trig:
    def __init__(self,args):
        self.args = args
        self.cartesian = False
        self.exp= False

    def change_to_exp(self):
        self.exp= True

    def compute(self):
        for i,arg in enumerate(self.args):
            self.args[i] = sym.simplify(arg)

    def convert_to_cartesian(self):
        self.cartesian = True

    def convert_to_polar(self):
        self.cartesian = False

class Sin(Trig):
    def __init__(self,args):
        Trig.__init__(self,args)

    def symcomp(self):
        #if len(self.args) > 1:
        #    self.symb = '*sin(('+self.args[0]+')'+'*phi_alpha +('+self.args[1]+')*phi_beta)'
        #else:
        #    self.symb = '*sin(('+self.args[0]+')*phi)'
        self.symb = '*sin('
        self.symb = self.symb+self.args[0]+f'*phi_{1}'
        for i in range(1,len(self.args)):
            self.symb=self.symb+'+'+self.args[i]+f'*phi_{i+1}'
        self.symb=self.symb+')'

    def compile(self):
        if len(self.args) > 1:
            if self.cartesian != True:
                arginsin=self.args[0]*sym.Symbol(f'phi_{1}',real=True)
                for i in range(1,len(self.args)):
                    arginsin=arginsin+self.args[i]*sym.Symbol(f'phi_{i+1}',real=True)
                if (self.exp==True):
                    self.sympy_form = sym.exp(sym.I*arginsin)
                else:
                    self.sympy_form = sym.sin(arginsin)
                #self.sympy_form = sym.sin(self.args[0]*sym.Symbol('phi_alpha',real=True)+self.args[1]*sym.Symbol('phi_beta',real=True))
            else:
                arginsin=self.args[0]*sym.atan2(sym.Symbol(f'y_{1}',real=True),sym.Symbol(f'x_{1}',real=True))
                for i in range(1,len(self.args)):
                    arginsin=arginsin+self.args[i]*sym.atan2(sym.Symbol(f'y_{i+1}',real=True),sym.Symbol(f'x_{i+1}',real=True))
                expr=sym.sin(arginsin)
                #expr = sym.sin(self.args[0]*sym.atan2(sym.Symbol('y_alpha',real=True),sym.Symbol('x_alpha',real=True))+self.args[1]*sym.atan2(sym.Symbol('y_beta',real=True),sym.Symbol('x_beta',real=True)))
                self.sympy_form = sym.factor(sym.expand(expr,trig=True)) #sym.factor
        else:
            if self.cartesian != True:
                if (self.exp==True):
                    self.sympy_form = sym.exp(-sym.I*self.args[0]*sym.Symbol(f'phi_{1}',real=True))
                else:
                    self.sympy_form = sym.sin(self.args[0]*sym.Symbol(f'phi_{1}',real=True))
            else:
                expr = sym.sin(self.args[0]*sym.atan2(sym.Symbol('y_1',real=True),sym.Symbol('x_1',real=True)))
                self.sympy_form = sym.factor(sym.expand(expr,trig=True)) #sym.factor

class Cos(Trig):
    def __init__(self,args):
        Trig.__init__(self,args)

    def symcomp(self):
        #if len(self.args) > 1:
        #    self.symb = '*cos(('+self.args[0]+')'+'*phi_alpha +('+self.args[1]+')*phi_beta)'
        #else:
        #    self.symb = '*cos(('+self.args[0]+')*phi)'
        self.symb = '*cos('
        self.symb = self.symb+self.args[0]+f'*phi_{1}'
        for i in range(1,len(self.args)):
            self.symb=self.symb+'+'+self.args[i]+f'*phi_{i+1}'
        self.symb=self.symb+')'
            
    def compile(self):
        if len(self.args) > 1:
            if self.cartesian != True:
                argincos=self.args[0]*sym.Symbol(f'phi_{1}',real=True)
                for i in range(1,len(self.args)):
                    argincos=argincos+self.args[i]*sym.Symbol(f'phi_{i+1}',real=True)
                if (self.exp==True):
                    self.sympy_form = sym.exp(sym.I*argincos)
                else:
                    self.sympy_form = sym.cos(argincos)
                #self.sympy_form = sym.cos(self.args[0]*sym.Symbol('phi_alpha',real=True)+self.args[1]*sym.Symbol('phi_beta',real=True))
            else:
                argincos=self.args[0]*sym.atan2(sym.Symbol(f'y_{1}',real=True),sym.Symbol(f'x_{1}',real=True))
                for i in range(1,len(self.args)):
                    argincos=argincos+self.args[i]*sym.atan2(sym.Symbol(f'y_{i+1}',real=True),sym.Symbol(f'x_{i+1}',real=True))
                expr=sym.cos(argincos)
                #expr = sym.cos(self.args[0]*sym.atan2(sym.Symbol('y_alpha',real=True),sym.Symbol('x_alpha',real=True))+self.args[1]*sym.atan2(sym.Symbol('y_beta',real=True),sym.Symbol('x_beta',real=True)))
                self.sympy_form = sym.factor(sym.expand(expr,trig=True)) #sym.factor
        else:
            if self.cartesian != True:
                if (self.exp==True):
                    self.sympy_form = sym.exp(sym.I*self.args[0]*sym.Symbol(f'phi_{1}',real=True))
                else:
                    self.sympy_form = sym.cos(self.args[0]*sym.Symbol(f'phi_{1}',real=True))
            else:
                expr = sym.cos(self.args[0]*sym.atan2(sym.Symbol('y_1',real=True),sym.Symbol('x_1',real=True)))
                self.sympy_form = sym.factor(sym.expand(expr,trig=True)) #sym.factor







def requires_signswap(rotational_eigenval):
    rot_eigenval_arg = copy(rotational_eigenval)
    if sym.im(rot_eigenval_arg).is_negative:
        return True
    else:
        return False    

from vhegen.tables.formulas import requirements_dct, return_formula

def get_root_formula(eigenvals,modes,symmetry):
    #symmetry symmetry dictionary
    #n_arg rotation number
    #modes list of vibrational modes
    #eigenvals, eigenvals from one or two electronic states
    #print(modes)
    #print(eigenvals)
    n_arg= symmetry['rot']
    n = str(n_arg)
    rotational_eigenval = eigenvals[0]
    refl_Re = eigenvals[1]
    refl_Im = eigenvals[2]
    inver=eigenvals[3]
    searchmodes = [mode[0] for mode in modes]

    #if len(modes) < 2:
    #    searchmodes.append('A')
    #    searchmodes = reorder(searchmodes)
    #if requires_signswap(rotational_eigenval) == True:
    #    reqs = [sym.re(rotational_eigenval) - sym.im(rotational_eigenval)*1j,searchmodes]
    #else:
    #    reqs = [rotational_eigenval,searchmodes]
    formula=return_formula(n,modes,symmetry)
    return(formula)
    #for k in requirements_dct:
    #    #print(reqs)
    #    #print(requirements_dct[k])
    #    if requirements_dct[k] == reqs:
    #        formula = return_formula(n,k)
    #        #print(formula)
    #        #exit()
    #        if requires_signswap == True:
    #            for term in formula:
    #                print(term)
    #                if isinstance(term,ET):
    #                    term.prefactor = sym.conjugate(term.prefactor)     
    #        #for term in formula:
    #        #    print(term.indices)     
    #        #exit()
    #        if refl_Im == 0:
    #            max_c = len(formula) - 1
    #            c = 0
    #            while c <= max_c:
    #                if isinstance(formula[c],Term):
    #                    if type(formula[c].prefactor) == complex:
    #                        del formula[c]
    #                        c -= 1
    #                        max_c -= 1
    #                c += 1
    #        return(formula)
    raise Exception('VHEGENError: Could not find root formula for rotational eigenvalue '+str(rotational_eigenval)+' and modes '+str(modes)+'.')

def read_index_arg(expr):
    #expr one type of index
    step = 1
    if expr[0] == 'K':
        step=2
    return step

def pos_ints(max,step):
    return list(range(0,max+1,step))

def any_ints(max):
    return list(range(-max,max+1))

def gen_index_attrs(expr,max_val):
    #expr: one index in list of indices
    #max_val: order of expansion (i.e. sum of indice values)
    step= read_index_arg(expr)
    if 'M' in expr:
        vals = any_ints(max_val)
    else:
        vals = pos_ints(max_val,step)

    return vals

def gen_index_lists(indices,order,eigenvals,nrot):
    val_list = []
    native_par_list = []
    for i in indices: #regular case
            vals = gen_index_attrs(i,order)
            val_list.append(vals)

    if eigenvals[2] == 0: #take real case
        m_found = False
        for c,i in enumerate(indices):
            if 'M' in i:
                m_found = True
                m_index = c #only need to keep first m positive
                break

        if eigenvals[0] == 0:
            if m_found == True:
                c = 0
                while c <= (len(val_list[m_index]) - 1):
                    if val_list[m_index][c] < 0:
                        del val_list[m_index][c]
                        c -= 1
                    c += 1
        elif abs(eigenvals[0]) == nrot//2:
            if m_found == True:
                c = 0
                while c <= (len(val_list[m_index]) - 1):
                    if val_list[m_index][c] < 0:
                        del val_list[m_index][c]
                        c -= 1
                    c += 1
        #print(eigenvals)
        #print(val_list)
        #exit()
    return val_list

def gen_real_restraint(list_of_terms,indices,order,eigenvals,nrot):

    if eigenvals[2] == 0: #take real case
        for c,i in enumerate(indices):
            #print(list_of_terms)
            #print(i,list_of_terms[c])
            if 'M' in i and list_of_terms[c]<0:
                return False
            if 'M' in i and list_of_terms[c]>0:
                return True
    return True
        
def adapt_to_unimodal(formula,modes):
    adapted_formula = []
    if formula != [0]:
        for term in formula:
            term.adapt_to_unimodal(modes)
            adapted_formula.append(term)
        
    return adapted_formula

def get_sym_props(pointgroup):
    sym_props = {'inver': False,
                 'refl': False,
                 'hrefl':False}
    if ('D' in pointgroup) or ('V' in pointgroup):
        sym_props['refl'] = True
    if 'H' in pointgroup:
        if '4' in pointgroup:
            sym_props['inver'] = True
        elif '3' in pointgroup:
            sym_props['hrefl'] = True
    return sym_props

def get_constraint(eigenvals, vib_modes, operation):
    princ_rot = eigenvals[0]
    refl_Re, refl_Im = eigenvals[1], eigenvals[2]
    inver_eigenval = eigenvals[3]
    try:
        os.chdir(os.getcwd() + '/constraints')
        constraintfiles = os.listdir(os.getcwd())
        constraintargs = [sym.sympify(i[0:i.index('_')].replace('X','*')) for i in constraintfiles]
        #Open refl.sym file
        for arg in constraintargs:
            if sym.simplify(princ_rot-arg) == 0:
                with open(replace_all(str(arg),{'*':'X',' ':''})+"_"+operation+".sym","r") as sym_file:
                    constraint_lines  = sym_file.readlines()
            elif sym.simplify(sym.re(princ_rot)-sym.im(princ_rot)*1j - arg) == 0:
                with open(replace_all(str(arg),{'*':'X',' ':''})+"_"+operation+".sym","r") as sym_file:
                    constraint_lines  = sym_file.readlines()
        os.chdir('..')
    except OSError as e:
        print(e)
        
    vib_modes = [i.replace("''",'"') for i in vib_modes]

    if operation == 'refl':
        eigen_req = '[' + str(refl_Re) + ',' + str(refl_Im) + ']'
        vib_modes = [replace_all(i,{'G':'','U':'',"'":'','"':''}) for i in vib_modes]

    elif operation == 'inver' or operation == 'hrefl':
        eigen_req = '[' + str(inver_eigenval) + ']'
        vib_modes = [replace_all(i,{'1':'','2':''}) for i in vib_modes]

    if len(vib_modes) == 1:        
        if operation == 'refl':
            vib_modes.append('A1')
        elif operation == 'inver':
            vib_modes.append('AG')
        elif operation == 'hrefl':
            vib_modes.append("A'")
    mode_req = vib_modes
    max_count = len(constraint_lines) - 1
    for count,line in enumerate(constraint_lines):
        modes_in_line = ''.join(line.split(',')[0])
        modes_in_line = modes_in_line[1:-1]
        modes_in_line = modes_in_line.split('.')
        line = line.replace('\n','')  
        if set([str(mode) for mode in mode_req]) == set(modes_in_line) and eigen_req in line:
            match_line = line
        if (count == max_count):
            try:
                match_line
            except NameError: #no matching constraints
                return {}
    constraints = match_line.split(': ',1)[1]

    constraints = constraints.split(',')
    constraints_dict = {}
    for constraint in constraints:
        index = constraint.split(' ')[0]
        restriction = ' '.join(constraint.split(' ')[1:])
        constraints_dict[index] = restriction
    if 'all' in constraints_dict:
        if constraints_dict['all'] == 'nr':
            del constraints_dict['all'] #check if any composite cases where 
    return constraints_dict
    
def load_constraints(sym, eigenvals, vib_modes):
    sym_props = get_sym_props(sym)
    matrix_element_constraints = {}
    for e in eigenvals:
        constraints_dict = {}
        if sym_props['refl'] == True:
            constraints_refl = get_constraint(eigenvals[e], vib_modes, 'refl')
            for key in constraints_refl:
                if key not in constraints_dict:
                    constraints_dict[key] = constraints_refl[key]
                else:
                    constraints_dict[key] += constraints_refl[key]
        if sym_props['inver'] == True:
            constraints_inver = get_constraint(eigenvals[e], vib_modes, 'inver')
            for key in constraints_inver:
                if key not in constraints_dict:
                    constraints_dict[key] = constraints_inver[key]
                else:
                    constraints_dict[key] += constraints_inver[key]
        if sym_props['hrefl'] == True:
            constraints_hrefl = get_constraint(eigenvals[e], vib_modes, 'hrefl')
            for key in constraints_hrefl:
                if key not in constraints_dict:
                    constraints_dict[key] = constraints_hrefl[key]
                else:
                    constraints_dict[key] += constraints_hrefl[key]
        matrix_element_constraints[e] = constraints_dict
    return matrix_element_constraints

def apply_constraints(local_constraints, p, fitted_term,unfitted_term): #returns whether constraints are sat (True) or not sat (False)
    sat_constraints = 0
    if local_constraints != {}:
        if 'all' in local_constraints.keys():
            if 'na' in local_constraints['all']: #all na case
                return False #return not sat
            else:
                if len(local_constraints) == 1: #all nr case. len 1 check ensures theres no other constraints to consider.
                    return True #return sat

        for c in local_constraints:
            #strip whitespace escape sequences
            if isinstance(local_constraints[c],set):
                local_constraints[c] = set([i.strip() for i in local_constraints[c]])
            else:
                local_constraints[c] = local_constraints[c].strip()
            
            #local_constraints[c] = local_constraints[c].strip() #remove whitespace 
            if type(c) == int: #handle summing-index based constraints
                if type(local_constraints[c]) == set:
                    len_cond = len(local_constraints[c])
                    met_cond = 0
                    for i in local_constraints[c]:
                        if constraint_funcs[i](c,p) == True:
                            met_cond += 1
                    if met_cond == len_cond:
                        sat_constraints += 1
                elif constraint_funcs[local_constraints[c]](c,p) == True:
                    sat_constraints += 1

            elif type(c.split('&')[0]) == int: #handle pairwise summing-index based constraints
                if constraint_funcs[local_constraints[c]]([int(i) for i in c.split('&')],p) == True:
                    sat_constraints += 1

            elif 'nz' in local_constraints[c]: #handle nz constraints
                if 'if' in local_constraints[c]: #has condition
                    cond = (local_constraints[c].split(' '))[-2:]
                else:
                    cond = []
                if '&' in c: #pairwise nz
                    c1 = c.split('&')[0]
                    c2 = c.split('&')[1]
                    if nz(c1,cond,fitted_term,unfitted_term,p) == True or nz(c2,cond,fitted_term,unfitted_term,p) == True:
                        sat_constraints +=1
                else:
                    if nz(c,cond,fitted_term,unfitted_term,p) == True:
                        sat_constraints += 1
    return sat_constraints == len(local_constraints)

#Constraint functions
def even(i,p):
    if p[i] % 2 != 0:
        return False
    else:
        return True

def odd(i,p):
    if p[i] % 2 == 0:
        return False
    else:
        return True

def ee_or_oo(i,p):
    if (p[i[0]] % 2 + p[i[1]] % 2) == 1:
        return False
    else:
        return True

def eo_or_oe(i,p):
    if (p[i[0]] % 2 + p[i[1]] % 2) != 1:
        return False
    else:
        return True

def non_neg(i,p):
    if p[i] >= 0:
        return True
    else:
        return False

def nm_postproc(i,p):
    if p[i][0] >  0:
        return True 
    if p[i][0] == 0 and p[i][1] >= 0:
        return True
    else:
        return False 

def na(i,p):
    return False

def nr(i,p):
    return True

def nz(arg,cond,fitted_term,unfitted_term,p): #coeff nz and trig nz constraints
    meets_req = False
    satisfy_arg = False
    if cond == []:
        meets_cond = True
    else: #if condition isn't true, constraint need not be applied.
        for c,i in enumerate(unfitted_term.indices):
            if cond[0] in i:
                cond_index = c
        if cond[1] == 'even':
            if p[cond_index] % 2 == 0:
                meets_cond = True
            else:
                meets_cond = False 
        else: #odd
            if p[cond_index] % 2 != 0:
                meets_cond = True
            else:
                meets_cond = False
    if meets_cond == False: #bypass
        meets_req = True
        return meets_req

    else: #nz constrain
        if arg in ['sin','cos']: #handle trig nz
            for part in fitted_term.funcs:
                if (arg == 'sin') and (isinstance(part,Sin)):
                    satisfy_arg = True
                elif (arg == 'cos') and (isinstance(part,Cos)):
                    satisfy_arg = True
        else: #handle coeff nz 
            satisfy_arg = coeff_nz(arg,fitted_term)
    if satisfy_arg == True:
        meets_req = True
    else:
        meets_req = False
    return meets_req

def coeff_nz(arg,term):
    validate = 0
    if arg[1] != '%':
        if term.coeff.stem == arg[1]:
            validate += 1
    else:
        validate += 1
    if not arg.endswith(R'%%'):
        index_parities = [arg[2],arg[3]]
        for c,i in enumerate(index_parities):
            if i == 'e':
                if int(term.indices[c]) % 2 == 0:
                    validate += 1
            elif i == 'o':
                if int(term.indices[c]) % 2 != 0:
                    validate += 1
    else:
        validate += 2
    if validate == 3:
        return True
    else:
        return False

def get_symbolic_formula(formula):
    symb_formula = ''
    if formula != [0]:
        for c,term_formula in enumerate(formula):
            symb_term_formula = term_formula.compile_formula()
            if symb_term_formula[1] != '-' and c != 0:
                symb_formula += '+'
            symb_formula += term_formula.compile_formula()
    return symb_formula

def get_matrix_element_expansion(formula,order,e_coord_system,eigenvals,symmetry,modes):
    
    expansions = []
    params = []
    
    if e_coord_system == 'both':
        iterates = 2
    else:
        iterates = 1
    for i in range(iterates):
        if (i==0):
            print('generating polar expansion')
        else:
            print('generating cartesian expansion')
        total_expansion = []
        if formula != [0]:
            for f,term_formula in enumerate(formula):
                #term_formula.grab_constraints(constraints)
                if e_coord_system == 'pol' or i == 0:
                    term_formula.convert_to_polar()
                if e_coord_system == 'cart' or i == 1:
                    term_formula.convert_to_cartesian()
                #if e_coord_system == 'expandedpol' or i == 2:
                #    term_formula.convert_to_expandedpol()
                #term_formula.expand(order,eigenvals,symmetry,modes)
                print(f'generating expansion for term {f}')
                term_formula.expand_new(order,eigenvals,symmetry,modes)
                
                #print(term_formula.expansions)   
                #print(term_formula.parameters)   
            for f,term_formula in enumerate(formula):
                if e_coord_system=='pol' or i==0:
                    if (len(formula)==4 and f<2):
                        for pos1 in range(len(term_formula.expansions)):
                            #print(pos1)
                            #print(term_formula.indiceslist[pos1])
                            if term_formula.indiceslist[pos1] in formula[f+2].indiceslist:
                                term_formula.expansions[pos1].change_to_exp()
                                pos2=formula[f+2].indiceslist.index(term_formula.indiceslist[pos1])
                            #    #print(term_formula.expansions[pos1])
                            #    #print(formula[f+2].expansions[pos2])
                            #    #term_formula.expansions[pos1]=(term_formula.expansions[pos1]+formula[f+2].expansions[pos2]).simplify()
                                del formula[f+2].expansions[pos2]#.prefactor=sym.S.Zero
                                del formula[f+2].indiceslist[pos2]
                            #    #print(term_formula.expansions[pos1])
                print(f'compiling term {f}')
                term_formula.compile_expansions()
                total_expansion += term_formula.expansions
                params += term_formula.parameters
        expansions.append(total_expansion)
        #print(' ')
    params = set(params)
    built_expansions = [0]*iterates
    for c,i in enumerate(expansions):
        for exp in i:
            built_expansions[c] += exp
    
    return built_expansions,params

constraint_funcs = {'even':even,
                    'odd':odd,
                    'ee_or_oo':ee_or_oo,
                    'eo_or_oe':eo_or_oe,
                    'non_neg': non_neg,
                    'nm_postproc': nm_postproc,
                    'na':na,
                    'nr':nr}

#EOF


from sympy import Matrix, Symbol, Add, Mul, UnevaluatedExpr, conjugate, re, im, printing, simplify, sqrt, sympify, expand_complex
from sympy.physics.quantum import Ket, Bra
from vhegen.modules.glbls import state_components_dct, replace_all
from copy import copy,deepcopy

def kdelta(p,q):
    if p==q:
        return 1
    else:
        return 0

def refl_parity(state):
    if ('2' in state):
        return 2
    elif ('1' in state):
        return 1
    else:
        return 0

def inver_parity(state):
    if ('U' in state) or ("''" in state):
        return -1
    elif ('G' in state) or ("'" in state):
        return 1
    else:
        return 0

def get_state_components(states):

    label_dct = {0: R'_\alpha',
                 1: R'_\beta'}

    labelling = False
    state_types = [s[0] for s in states]
    if len(state_types) == 2:
        if state_types[0] == state_types[1]: #Gamma + Gamma 
            labelling = True

    state_components = []
    for c,s in enumerate(state_types):
        if labelling == True:
            state_components += [s_i+label_dct[c] for s_i in state_components_dct[s]]
        else:
            state_components += state_components_dct[s]
    
    return state_components
            
class VibronicMatrix:
    def __init__(self,symmetry,states):

        from vhegen.tables.matrices import requirements_dct, matrix_dct, return_matrices
        from vhegen.tables.eigenvalues import return_eigenvals

        self.original_mat, self.symmetrized_mat, self.transform_mat = return_matrices(symmetry,states)
        self.symmetry_eigenvals = return_eigenvals(symmetry,states) #Matrix element symmetry eigenvalues
        #print(self.symmetry_eigenvals)
        #exit()
        self.state_components = get_state_components(states) #Matrix ket-bra basis representation
        self.get_matrix_product_form()

    def get_matrix_product_form(self):
        self.ket_row = Matrix([Ket(s) for s in self.state_components]).transpose()
        self.bra_col = Matrix([Bra(s) for s in self.state_components])
        self.form = UnevaluatedExpr(self.ket_row)*UnevaluatedExpr(self.symmetrized_mat)*UnevaluatedExpr(self.bra_col)
        return self.form

    def __str__(self):
        return str(self.symmetrized_mat)

    def get_dependencies(self):

        independent_set = set()
        for m_e in self.symmetrized_mat:
            if not isinstance(m_e,conjugate) and m_e != 0:
                independent_set.add(m_e)

        self.dependencies = {}
        for indep_e in independent_set:
            self.dependencies[indep_e] = []
            for c,e in enumerate(self.symmetrized_mat):
                if e == indep_e:
                    self.dependencies[indep_e].append(self.original_mat[c])
                if conjugate(e) == indep_e:
                    self.dependencies[indep_e].append(conjugate(self.original_mat[c]))

        return self.dependencies

    def format_TeX_mat(self):
        tex_mat = deepcopy(self.original_mat)
        for c,i in enumerate(tex_mat):
            tex_mat[c] = format_matrix_element(i)

        tex_form = UnevaluatedExpr(self.ket_row)*UnevaluatedExpr(tex_mat)*UnevaluatedExpr(self.bra_col)
        tex_formatted_form = R'$\hat{H}='+printing.latex(tex_form,mat_delim='(',mat_str='matrix')+'$'
            
        return tex_formatted_form

    def returnpos(self):
        self.element_position={}
        scrublist=[]
        l=len(self.original_mat)
        for i in range(1,5):
            if (i*i==l):
                l=i
                break
        self.element_position['matsize']=l
        for c,e in enumerate(self.original_mat):
            if e != 0:
                print(c)
                print(e)
                cpos=(c)//l+1
                rpos=(c-(cpos-1)*l)+1
                pos=[cpos,rpos]
                if e not in scrublist:
                    scrublist.append(e)
                self.element_position[e]=pos
        #print(self.element_position)

    def change_basis(self): #Complex to real matrix transformation

        symmetrized_mat_expanded = deepcopy(self.symmetrized_mat)
        for c,e in enumerate(symmetrized_mat_expanded): #Expand all elements into real and imaginary parts
            if type(e) == conjugate:
                symmetrized_mat_expanded[c] = re(conjugate(e)) - 1j*im(conjugate(e))
            else:
                symmetrized_mat_expanded[c] = re(e) + 1j*im(e)

        evolved_mat = self.transform_mat*symmetrized_mat_expanded*(self.transform_mat.H)

        #Update matrix product form:
        for c in range(0,len(self.state_components)):
            self.state_components[c] = replace_all(self.state_components[c], {'+':'X','-':'Y'})
        
        self.element_position={}
        scrublist=[]
        l=len(self.original_mat)
        for i in range(1,5):
            if (i*i==l):
                l=i
                break
        self.element_position['matsize']=l
        #print('converting')
        for c,e in enumerate(self.original_mat):
            evolved_mat[c] = simplify(expand_complex(evolved_mat[c])) #Simplify evolved mat element-wise
            if e != 0:
                e = replace_all(str(e), {'+':'X','-':'Y'})
                #print(c)
                cpos=(c)//l+1
                rpos=(c-(cpos-1)*l)+1
                pos=[cpos,rpos]
                if e not in scrublist:
                    scrublist.append(e)
                self.element_position[Symbol(e)]=pos
                self.original_mat[c] = Symbol(e)

        #print(self.element_position)
        #print(self.original_mat)
        self.evolved_mat = evolved_mat
        self.get_matrix_product_form()

        #Element mapping
        mapping = {}
        for c,e in enumerate(self.original_mat):
            if e != 0:
                e_str = replace_all(str(e), {'+':'X','-':'Y'})
                mapping[Symbol(e_str)] = self.evolved_mat[c]

        return mapping
        
def get_dependent_elements(expansion_dct,count,obj):
    #expansion_dct: dictionary of expansions with label o for order
    #count is dictionary of number of terms for each order o
    #obj vhegen class
    print('\nApplying expansion dependencies...\n')
    all_expansions_dct = {} 
    dependent_count = {}
    for o in expansion_dct: #for order
        all_expansions_dct[o] = {}
        dependent_count[o] = {}
        for e in expansion_dct[o]:
            for dep_e in obj.matrix.dependencies[e]:
                if isinstance(dep_e,conjugate):
                    all_expansions_dct[o][conjugate(dep_e)] = [conjugate(expansion_dct[o][e][c]) for c,i in enumerate(expansion_dct[o][e])]
                    dependent_count[o][conjugate(dep_e)] = count[o][e]
                else:
                    all_expansions_dct[o][dep_e] = expansion_dct[o][e]
                    dependent_count[o][dep_e] = count[o][e]
    return all_expansions_dct

def format_matrix_element(m_e):
    if m_e != 0:
        if isinstance(m_e,conjugate):
            m_e_str = replace_all(str(conjugate(m_e)), {'__':'', 'alpha':R'{\alpha}', 'beta':R'{\beta}'})
            return (Symbol(R'{H_{'+str(m_e_str)+R'}}^*'))
        else:
            m_e_str = replace_all(str(m_e), {'__':'', 'alpha':R'{\alpha}', 'beta':R'{\beta}'})
            return Symbol(R'H_{'+m_e_str+R'}')
    else: return 0

def prune_dependent_elements(real_matrix_elements):
    indep_m_e = []
    m_e = [str(i) for i in real_matrix_elements]

    if 'X__A' in m_e or 'X__B':
        for i in m_e:
            if i[0] == 'X' or i[0] == 'Y':
                indep_m_e.append(Symbol(i))

    elif 'X__X' in m_e:
        for i in m_e:
            if i[0] == 'X':
                indep_m_e.append(Symbol(i))

    elif '__' in m_e[0]:
        m_e_lists = []
        for i in m_e:
            m_e_lists.append(i.split('__'))
        for i in m_e_lists:
            if 'alpha' in i[0] and 'X' in i[0]:
                indep_m_e.append(Symbol('__'.join(i)))

    
    else:
        indep_m_e = real_matrix_elements

    return indep_m_e

def map_elements(expansions, mapping):
    coords = 1
    if len(list(list(expansions.values())[0].values())[0]) > 1:
        coords = 2
    mapped_expansions = {}
    term_inheritance = {}
    independent_elements = prune_dependent_elements([i for i in mapping])
    for o in expansions:
        sub_tuples = [(e,expansions[o][e][0]) for e in expansions[o].keys()]
        if coords == 2:
            sub_tuples_2 = [(e,expansions[o][e][1]) for e in expansions[o].keys()]
        mapped_expansions[o] = {}
        term_inheritance[o] = {}
        for k in mapping:
            get_inheritance = False
            if k in independent_elements:
                get_inheritance = True
                inheritance_count = {}
            expanded_terms = []
            if type(mapping[k]) == Add:
                k_args = list(mapping[k].args)
            else:
                k_args = [mapping[k]]
            k_args = (k_args, [i.free_symbols for i in k_args])
            for c,element in enumerate(k_args[1]):
                for c2,item in enumerate(sub_tuples):
                    if item[0].free_symbols == element:
                        if coords == 1:
                            substituted_args = [k_args[0][c].subs(item[0],item[1])]
                        else:
                            substituted_args = [k_args[0][c].subs(item[0],item[1]),k_args[0][c].subs(sub_tuples_2[c2][0],sub_tuples_2[c2][1])]
                        expanded_terms.append(substituted_args)
                        term_arguments = substituted_args[0]

                        if type(term_arguments) == Symbol:
                            term_count = 1
                        elif type(term_arguments) == Mul: #handle numeric prefactor and single terms
                            sum_exists = False
                            for part in term_arguments.args:
                                if type(part) == Add:
                                    sum_exists = True
                                    term_count = len(part.args)
                            if sum_exists == False:
                                term_count = 1
                        else: #sym.Add
                            try:
                                term_count = len(substituted_args[0].args)
                            except AttributeError:
                                term_count = 0
                        if get_inheritance == True:
                            inheritance_count[k_args[1][c].pop()] = term_count

            sum_terms = 0
            for term in expanded_terms:
                sum_terms += term[0]
            if coords == 2:
                sum_terms2 = 0
                for term in expanded_terms:
                    sum_terms2 += term[1]
                sum_terms = [sum_terms,sum_terms2]
            else:
                sum_terms = [sum_terms]
            mapped_expansions[o][k] = sum_terms
            if get_inheritance == True:
                term_inheritance[o][k] = inheritance_count
    return mapped_expansions, term_inheritance
    
#EOF



#VHEGEN testrun script

import vhegen as VHE
from time import time

problem = VHE.modules.input.prepare_input('D6h',"E1g","e2g+e2g+e1u+e1g+e2u",'0,6','benzene')
t1 = time()

vhegen_instance = VHE.VHEGEN(problem)

print(vhegen_instance.return_init())

vhegen_instance.set_e_coordinates('both')

vhegen_instance.set_basis('both')

vhegen_instance.auto()

vhegen_instance.pdflatex()

vhegen_instance.mctdhoutput()

print('\nJob complete after '+str(round(time()-t1,3))+' seconds.')
print('\nTestrun complete without errors.')



import sympy as sym
from sympy.printing.latex import LatexPrinter, print_latex
import os
import subprocess
import vhegen.modules.electronic as el

log_str = ''

def log_append(string):
    #add contents to log file
    global log_str
    log_str += str(string) + '\n'

def format_mulliken_scripts(irrep,t):
    if t == 's':
        irrep_str = irrep[0].upper()
    else:
        irrep_str = irrep[0].lower()
    primes = 0
    for char in irrep:
        if char == "'":
            primes += 1
    if primes == 0:
        irrep_str += '_{' + irrep[1:].lower() + '}'
    else:
        irrep_str += primes*"'" + '_{' + irrep[1:].replace("'","") + '}'
    return irrep_str

def problem_format2TeX(states,modes):
    problem_str = '$'
    if len(states) > 1:
        problem_str += '('
    for c,s in enumerate(states):
        if c>0:
            problem_str +='+'
        problem_str += format_mulliken_scripts(s,'s')
    if len(states) > 1:
        problem_str += ')'
    problem_str += R' \otimes '
    if len(modes) > 1:
        problem_str += '('
    for c2,m in enumerate(modes):
        if c2>0:
            problem_str+='+'
        problem_str += format_mulliken_scripts(m,'m')
    if len(modes) > 1:
        problem_str += ')'
    problem_str += '$'
    return problem_str

def prune_dependent_terms(count):
  keys = list(count.keys())
  unique_keys = []
  unique_count = {}
  for i in keys:
    if sym.Symbol(str(i)[::-1]) not in unique_keys:
      unique_keys.append(i)
  for k in unique_keys:
    unique_count[k] = count[k]
  return(unique_count)

def count_format2TeX(count): 
  #{+A: 8, -A: 8, A+: 8, A-: 8}
  count = prune_dependent_terms(count)
  count_str = ''
  for c,m_e in enumerate(count):
    count_str += '$'+str(el.format_matrix_element(m_e))+'$: ' + '$'+str(count[m_e]) +'$'
    if c != len(count)-1:
      count_str += ', '
    else:
      count_str += '.'
  return count_str

def realcount_format2TeX(count): #includes inheritance from complex
  #{XA: {+A: 6, ++: 10}, YA: {+A: 6}, AX: {+A: 6}, AY: {+A: 6}}
  count = prune_dependent_terms(count)
  count_str = ''
  for c,m_e in enumerate(count):
    sumval = sum(count[m_e][i] for i in count[m_e])
    count_str += '$'+str(el.format_matrix_element(m_e))+'$: ' + '$'+str(sumval) +'$'
    if sumval != 0:
      count_str += ' ('
      for c2,inherited_term in enumerate(count[m_e]):
        added= False
        if sumval == count[m_e][inherited_term]:
          count_str += 'all from $'+str(el.format_matrix_element(inherited_term))+'$)'
          break
        else:
          if count[m_e][inherited_term] != 0:
            count_str += '$'+str(count[m_e][inherited_term])+'$ from $'+str(el.format_matrix_element(inherited_term))+'$'
            added=True

        if c2 != len(count[m_e])-1:
          if added==True:
            count_str += ', '
        else:
          count_str += ')'

    if c!= len(count)-1:
      count_str += ', '
    else:
      count_str += '.'
  return count_str 

def compose_TeX_both_bases(sym,modes,TeX_matrix_complex,TeX_matrix_real,TeX_problem,TeX_expansions_complex,TeX_expansions_real):
    #print(TeX_expansions_complex)
    from vhegen.modules.input import return_acceptable_states
    amodes=return_acceptable_states(sym)
    modelabels=[]
    for i in range(len(modes)):
       for j in range(len(amodes.full)):
          if modes[i]==amodes.full[j]:
             modelabels.append(amodes.labels[j])
    posa=[]
    numa=0
    posb=[]
    numb=0
    pose=[]
    nume=0
    modes0=[mode[0] for mode in modes]
    for i in range(len(modes)):
        if modes0[i]=='E':
            pose.append(i)
            nume=nume+1
        if modes0[i]=='A':
            posa.append(i)
            numa=numa+1
        if modes0[i]=='B':
            posb.append(i)
            numb=numb+1
    indices=[]
    for i in range(numa):
        indices.append(Rf'z_{i+1} &\rightarrow& '+format_mulliken_scripts(modes[posa[i]],'m')+'\\\\ \n')
    for i in range(numb):
        indices.append(Rf'w_{i+1} &\rightarrow& '+format_mulliken_scripts(modes[posb[i]],'m')+'\\\\ \n')
    for i in range(nume):
        indices.append(Rf'\rho_{i+1},\phi_{i+1},x_{i+1},y_{i+1} &\rightarrow& '+format_mulliken_scripts(modes[pose[i]],'m')+'\\\\ \n')
        #indices.append(Rf'\phi_{i+1} &\rightarrow& '+format_mulliken_scripts(modes[pose[i]],'m')+'\\\\ \n')
    latexscript=R'\begin{eqnarray*}'+'\n'
    for i in range(len(indices)):
        latexscript=latexscript+indices[i]
    latexscript=latexscript+'\end{eqnarray*}'
    #print(latexscript)
    #exit()
    TeX_str = (R'\batchmode'+'\n'
           R'\documentclass[fleqn]{article}'+'\n'
           R'\usepackage{amsmath}'+'\n'
           #R'\usepackage{breqn}'+'\n'
           R'\usepackage{hyperref}'+'\n'
           R'\usepackage[margin=0.7in]{geometry}' + '\n'
           R'\setlength\mathindent{0pt}'+'\n'
           R'\allowdisplaybreaks'+'\n'
           R'\title{VHEGEN: A vibronic Hamiltonian expansion generator for trigonal and tetragonal polyatomic systems}'+'\n'
           R'\author{James Brown \and Robert A. Lang \and Riley J. Hickman \and Tao Zeng}'+'\n'
           R'\date{}'+'\n'
           R'\begin{document}'+'\n'
           R'\maketitle'+'\n'
           R'Thank you for using \texttt{VHEGEN}, the \texttt{V}-ibronic \texttt{H}-amiltonian \texttt{E}-xpansion \texttt{GEN}-erator for trigonal and tetragonal polyatomic systems. '
           R'This is a \texttt{VHEGEN} output file compiled by \texttt{pdflatex}. '
           R'If the \texttt{VHEGEN} package was used in research resulting in a publication, please reference the article in \textit{Computer Physics Communications} which describes the program ([doi here]). '
           R'Additional information regarding the matrix element expansion process, including the independent matrix element eigenvalues, their root formulas and constraints, and their transformation to the real basis (if applicable), can be found in the \texttt{log} output file. '
           R'For questions, bugs, or comments, please contact jbrown88@yorku.ca.\\\\'+'\n'
           R'\tableofcontents'+'\n'
           R'\newpage'+'\n'
           R'\section{Vibronic interaction}'+'\n'
           R''+TeX_problem+' in $'+sym['letter']+'_{'+sym['print'][1:].lower()+R'}$'+'\n'
           R''+latexscript+'\n'
           R'\section{Vibronic Hamiltonian operator in the complex $E$ basis}'+'\n'
           R''+TeX_matrix_complex.replace('__','')+'\n'
           R'\section{Matrix element expansions in the complex $E$ basis}'+'\n'
           R''+TeX_expansions_complex.replace('__','')+'\n'
           R'\section{Vibronic Hamiltonian operator in the real $E$ basis}'+'\n'
           R''+TeX_matrix_real.replace('__','')+'\n'
           R'\section{Matrix element expansions in the real $E$ basis}'+'\n'
           R''+TeX_expansions_real.replace('__','')+'\n'
           R'\end{document}')
    return TeX_str

def compose_TeX(sym,modes,TeX_matrix,TeX_problem,TeX_expansions):
    


    TeX_str = (R'\batchmode'+'\n'
               R'\documentclass[fleqn]{article}'+'\n'
               R'\usepackage{amsmath}'+'\n'
               #R'\usepackage{breqn}'+'\n'
               R'\usepackage[margin=0.7in]{geometry}' + '\n'
               R'\setlength\mathindent{0pt}'+'\n'
               R'\allowdisplaybreaks'+'\n'
               R'\usepackage{hyperref}'+'\n'
               R'\title{VHEGEN: A vibronic Hamiltonian expansion generator for trigonal and tetragonal polyatomic systems}'+'\n'
               R'\author{James Brown \and Robert A. Lang \and Riley J. Hickman \and Tao Zeng}'+'\n'
               R'\date{}'+'\n'
               R'\begin{document}'+'\n'
               R'\maketitle'+'\n'
               R'Thank you for using \texttt{VHEGEN}, the \texttt{V}-ibronic \texttt{H}-amiltonian \texttt{E}-xpansion \texttt{GEN}-erator for trigonal and tetragonal polyatomic systems. '
               R'This is a \texttt{VHEGEN} output file compiled by \texttt{pdflatex}. '
               R'If the \texttt{VHEGEN} package was used in research resulting in a publication, please reference the article in \textit{Computer Physics Communications} which describes the program. '
               R'Additional information regarding the matrix element expansion process, including the independent matrix element eigenvalues, their root formulas and constraints, and their transformation to the real basis (if applicable), can be found in the \texttt{log} output file. '
               R'For questions, bugs, or comments, please contact jbrown88@yorku.ca.\\\\'+'\n'
               R'\tableofcontents'+'\n'
               R'\newpage'+'\n'
               R'\section{Vibronic interaction}'+'\n'
               R''+TeX_problem+' in $'+sym['letter']+'_{'+sym['print'][1:].lower()+R'}$'+'\n'
               #R''+latexscript+'\n'
               R'\section{Vibronic Hamiltonian operator}'+'\n'
               R''+TeX_matrix.replace('__','')+'\n'
               R'\section{Matrix element expansions}'+'\n'
               R''+TeX_expansions.replace('__','')+'\n'
               R'\end{document}')
    return TeX_str

def format_constraint(constraint): #{'a%ee': 'nz','broe&bieo': 'nz'}
    string = ''
    if constraint == {}:
        return ''
    for c,con in enumerate(constraint):
        cons = con.split('&')
        formatted_cons = []
        for const in cons:
            formatted_con = const
            if 'nz' in constraint[con] and con not in ['cos','sin']:
                add = 0
                if const[1] != '%':
                    formatted_con = const[:1] + '^' + const[1:]
                    add += 1
                if const[2] != '%':
                    formatted_con = formatted_con[:(2+add)] + '_' + formatted_con[(add+2):]
            formatted_cons.append(formatted_con)
        formatted_constraints = ' '.join(formatted_cons)
        formatted_constraints = formatted_constraints.replace('%','')
        string += formatted_constraints +' '+ constraint[con]
        if c != len(constraint)-1:
            string += ', '
    return string

def format_constraints(constraints):
    string = 'Constraints:\n\n'
    for e in constraints:
        if constraints[e] == {}:
            formatted_constraint = 'all nr'
        else:
            formatted_constraint = format_constraint(constraints[e])
        string += 'H_'+str(e) +' : ' + formatted_constraint +'\n'
    string +='\n'
    return string
    
def new_file(filename,extension,path):
	
    if path.endswith('/') != True:
        if (os.path.exists(path)!=True):
            os.mkdir(path)
        path += '/'
    file = open(path+filename+'.'+extension, "w")
    return file

def TeX_write(filename,path,TeX_str):
    TeX_file = new_file(filename,'tex',path)
    TeX_file.write(TeX_str)
    TeX_file.close()

def mctdh_write(filename,path,mctdh_str,mctdh_str2):
    mctdh_file = new_file(filename,'op',path)
    mctdh_file.write(mctdh_str)
    mctdh_file.close()
    mctdh_file = new_file(filename,'inp',path)
    mctdh_file.write(mctdh_str2)
    mctdh_file.close()

def log_write(filename,path):
    #write log file to outputs
    global log_str
    log_file = new_file(filename,'log',path)
    log_file.write(log_str)
    log_file.close()

def convert_syntax(sympy_expr,syntax):
    if syntax.upper() not in ['LATEX','MATHEMATICA']:
        raise Exception('ConvertSyntaxError: could not recognize syntax.')
    else:
        if syntax.upper() == 'LATEX':
            return convert_to_LaTeX(sympy_expr)
        elif syntax.upper() == 'MATHEMATICA':
            return convert_to_Mathematica(sympy_expr)

def convert_to_MCTDH(sympy_expr,e,element_position,vhegenclass,e_coords):
    #print()
    #print(f'creating mctdh operator for matrix element {e}')
    #print(element_position[e])
    sympy_str=str(sympy_expr).replace('.0','').split(' ')
    if (e_coords=='cart'):
        if hasattr(vhegenclass, 'mctdhout')==False:
            vhegenclass.mctdhout=mctdh(vhegenclass.modes,e_coords,vhegenclass.filename)
        if (element_position[e][0]<=element_position[e][1]):
            vhegenclass.mctdhout.appendmctdh(sympy_str,element_position[e],e_coords)
    else:
        if hasattr(vhegenclass, 'mctdhoutp')==False:
            vhegenclass.mctdhoutp=mctdh(vhegenclass.modes,e_coords,vhegenclass.filename)
        if (element_position[e][0]<=element_position[e][1]):
            vhegenclass.mctdhoutp.appendmctdh(sympy_str,element_position[e],e_coords)

class MyLatexPrinter(LatexPrinter):
    """Print exp as \exp instead of e^.
    """
    
    
    _default_settings = {
        "full_prec": False,
        "fold_frac_powers": False,
        "fold_func_brackets": False,
        "fold_short_frac": True,
        "inv_trig_style": "abbreviated",
        "itex": False,
        "ln_notation": False,
        "long_frac_ratio": None,
        "mat_delim": "[",
        "mat_str": None,
        "mode": "plain",
        "mul_symbol": None,
        "order": None,
        "symbol_names": {},
        "root_notation": True,
        "mat_symbol_style": "plain",
        "imaginary_unit": "i",
        "gothic_re_im": False,
        "decimal_separator": "period",
        "perm_cyclic": True,
        "parenthesize_super": True,
        "min": None,
        "max": None,
    }  # type: Dict[str, Any]

    def _print_ExpBase(self, expr, exp=None):
        # TODO should exp_polar be printed differently?
        #      what about exp_polar(0), exp_polar(1)?
        tex = r"\exp \left(%s\right)" % self._print(expr.args[0])
        return self._do_exponent(tex, exp)

    def _print_Function(self, expr, exp=None):
        r'''
        Render functions to LaTeX, handling functions that LaTeX knows about
        e.g., sin, cos, ... by using the proper LaTeX command (\sin, \cos, ...).
        For single-letter function names, render them as regular LaTeX math
        symbols. For multi-letter function names that LaTeX does not know
        about, (e.g., Li, sech) use \operatorname{} so that the function name
        is rendered in Roman font and LaTeX handles spacing properly.

        expr is the expression involving the function
        exp is an exponent
        '''
        func = expr.func.__name__
        if hasattr(self, '_print_' + func) and \
                not isinstance(expr, AppliedUndef):
            return getattr(self, '_print_' + func)(expr, exp)
        else:
            args = [str(self._print(arg)) for arg in expr.args]
            # How inverse trig functions should be displayed, formats are:
            # abbreviated: asin, full: arcsin, power: sin^-1
            inv_trig_style = self._settings['inv_trig_style']
            # If we are dealing with a power-style inverse trig function
            inv_trig_power_case = False
            # If it is applicable to fold the argument brackets
            can_fold_brackets = self._settings['fold_func_brackets'] and \
                len(args) == 1 and \
                not self._needs_function_brackets(expr.args[0])

            inv_trig_table = [
                "asin", "acos", "atan",
                "acsc", "asec", "acot",
                "asinh", "acosh", "atanh",
                "acsch", "asech", "acoth",
            ]

            # If the function is an inverse trig function, handle the style
            if func in inv_trig_table:
                if inv_trig_style == "abbreviated":
                    pass
                elif inv_trig_style == "full":
                    func = "arc" + func[1:]
                elif inv_trig_style == "power":
                    func = func[1:]
                    inv_trig_power_case = True

                    # Can never fold brackets if we're raised to a power
                    if exp is not None:
                        can_fold_brackets = False

            if inv_trig_power_case:
                if func in accepted_latex_functions:
                    name = r"\%s^{-1}" % func
                else:
                    name = r"\operatorname{%s}^{-1}" % func
            elif exp is not None:
                func_tex = self._hprint_Function(func)
                func_tex = self.parenthesize_super(func_tex)
                name = r'%s^{%s}' % (func_tex, exp)
            else:
                name = self._hprint_Function(func)

            if can_fold_brackets:
                if func in accepted_latex_functions:
                    # Wrap argument safely to avoid parse-time conflicts
                    # with the function name itself
                    name += r" {%s}"
                else:
                    name += r"%s"
            else:
                name += r" \left(%s\right)"

            if inv_trig_power_case and exp is not None:
                name += r"^{%s}" % exp

            return name % ",".join(args)

def print_my_latex(expr):
    """ Most of the printers define their own wrappers for print().
    These wrappers usually take printer settings. Our printer does not have
    any settings.
    """
    print(MyLatexPrinter().doprint(expr))


def convert_to_LaTeX(sympy_expr):
    #print(sympy_expr)
    latex_expr = MyLatexPrinter().doprint(sympy_expr)#,fold_short_frac=True)
    #latex_expr =  sym.printing.latex(sympy_expr,fold_short_frac=True)
    latex_expr = latex_expr.replace('1.0','')
    latex_expr = latex_expr.replace('.0','')
    latex_expr = latex_expr.replace(R'\left(',R'(')
    latex_expr = latex_expr.replace(R'\right)',R')')
    latex_expr = add_dashes_to_latex(latex_expr)
    return latex_expr

def add_dashes_to_latex(latex_expr):
    tcount=0
    total_length=len(latex_expr)
    while tcount+200<total_length:
        tcount=tcount+200
        #go backwards to find split
        
        for i in range(200):
            tcount=tcount-1
            #print(tcount)
            #print(total_length)
            if (latex_expr[tcount]=='+' or latex_expr[tcount]=='-'):
                if (latex_expr[tcount-1]!=',' and latex_expr[tcount-1]!=R'{'):
                    #print(len(latex_expr))
                    latex_expr=latex_expr[:tcount+1]+R'\\'+'\n'+'&'+latex_expr[tcount+1:]
                    #print(latex_expr[tcount+1:tcount+8])
                    tcount=tcount+4
                    #print(len(latex_expr))
                    total_length+=4
                    break
    return latex_expr


def exec_pdflatex(filename,path):
    if not path.endswith('/'):
        path += '/'
    print('Sending output to '+path)
    if os.path.isfile(path+filename + '.pdf') == True:
        print('Overwriting '+ path+filename + '.pdf')
        os.remove(path+filename + '.pdf')
    print('Generating output PDF:\n')
    try:
        proc = subprocess.Popen(['pdflatex --interaction=batchmode -output-directory='+path, path+filename+'.tex'])
        proc.communicate()
        proc.communicate()
    except OSError:
        try:
            os.system('pdflatex --interaction=batchmode -output-directory='+path+' '+path+filename +'.tex')
            os.system('pdflatex --interaction=batchmode -output-directory='+path+' '+path+filename +'.tex') #double compile for ToC
        except OSError:
            print('Error making '+filename+'.pdf.\n')   
    '''Remove unwanted LaTeX output files .aux and .log'''
    if os.path.isfile(path+filename + '.aux') == True:
        os.remove(path+filename + '.aux')
    if os.path.isfile(path+filename + '.log') == True:
        os.remove(path+filename + '.log')

class mctdh:
    def __init__(self,modes,e_coords,filename):
        modes0=[mode[0] for mode in modes]
        nume=0
        numb=0
        numa=0
        tot=0
        for i in range(len(modes)):
            if modes0[i]=='E':
                nume=nume+1
            if modes0[i]=='A':
                numa=numa+1
            if modes0[i]=='B':
                numb=numb+1
        self.modeargs=[]
        self.massargs=[]
        self.masspos=[]
        for i in range(numa):
            self.modeargs.append(f'z{i+1}')
            self.massargs.append(f'mass_z{i+1}')
            self.masspos.append((tot))
            tot=tot+1
        for i in range(numb):
            self.modeargs.append(f'w{i+1}')
            self.massargs.append(f'mass_w{i+1}')
            self.masspos.append((tot))
            tot=tot+1
        for i in range(nume):
            self.massargs.append(f'mass_e{i+1}')
            if (e_coords=='cart'):
                self.modeargs.append(f'x{i+1}')
                self.modeargs.append(f'y{i+1}')
            else:
                self.modeargs.append(f'rho{i+1}')
                self.modeargs.append(f'phi{i+1}')
            self.masspos.append((tot,tot+1))
            tot=tot+2
        
        self.modeargs.append('el')
        self.nmodes=len(self.modeargs)
        self.dashes=''
        
        self.printmodes=(R' modes').ljust(25)+R' |'
        for j in range(27):
            self.dashes=self.dashes+R'-'
        for i in range(self.nmodes-1):
            if ('phi' in self.modeargs[i]):
                self.printmodes=self.printmodes+' '+self.modeargs[i].ljust(11)+R' |'
                for j in range(14):
                    self.dashes=self.dashes+R'-'
            else:
                self.printmodes=self.printmodes+' '+self.modeargs[i].ljust(4)+R' |'
                for j in range(7):
                    self.dashes=self.dashes+R'-'

        self.printmodes=self.printmodes+' '+self.modeargs[self.nmodes-1]
        for j in range(5):
            self.dashes=self.dashes+R'-'
        if (e_coords=='cart'):
            self.inpfile=(R'RUN-SECTION'+'\n'
                            R'name='+filename+'-cart'+'\n'
                            R'title='+filename+'-cart'+'\n'
                            R''+'\n'
                            R'OPERATOR-SECTION'+'\n'
                            R'opname='+filename+'-cart'+'\n'
                            R'end-operator-section'+'\n'
                            R''+'\n')
            self.title_section=(R'OP_DEFINE-SECTION'+'\n'
                            R'title'+'\n'
                            R'   '+filename+'-cart'+'\n'
                            R'end-title'+'\n'
                            R'end-op_define-section'+'\n'
                            R''+'\n')
        else:
            self.inpfile=(R'RUN-SECTION'+'\n'
                            R'name='+filename+'-pol'+'\n'
                            R'title='+filename+'-pol'+'\n'
                            R''+'\n'
                            R'OPERATOR-SECTION'+'\n'
                            R'opname='+filename+'-pol'+'\n'
                            R'end-operator-section'+'\n'
                            R''+'\n')
            self.title_section=(R'OP_DEFINE-SECTION'+'\n'
                            R'title'+'\n'
                            R'   '+filename+'-pol'+'\n'
                            R'end-title'+'\n'
                            R'end-op_define-section'+'\n'
                            R''+'\n')
        self.addpbasis()
        self.parameter_section=(R'PARAMETER-SECTION'+'\n')
        self.parameters_list=[]
        for i in range(len(self.massargs)):
            self.parameters_list.append(self.massargs[i])
        self.hamiltonian_section=(self.dashes+'\n'
                            R'HAMILTONIAN-SECTION'+'\n'
                            +self.dashes+'\n'
                            +self.printmodes+'\n'
                            +self.dashes+'\n')
        self.add_keo(e_coords)
        
        #self.modecombolist=[]
    
    def add_keo(self,e_coords):
        colsize=[]
        for j in range(self.nmodes):
            if ('phi' in self.modeargs[j]):
                colsize.append(12)
            else:
                colsize.append(5)
        for i in range(len(self.massargs)):
            if ('e' in self.massargs[i]): #need two or three lines for this mass
                if (e_coords=='cart'):
                    for j in range(2): #need two lines for each e-mode
                        self.hamiltonian_section+=(R'-1/2*'+self.massargs[i]).ljust(25)+R' |'
                        for k in range(self.nmodes-1): #check all modes for correct position
                            if (k!=self.masspos[i][j]):
                                self.hamiltonian_section+=(' 1').ljust(colsize[k])+R' |'
                            else:
                                self.hamiltonian_section+=' dq^2'.ljust(colsize[k])+R' |'
                        self.hamiltonian_section+=' 1'.ljust(colsize[k])+'\n'
                else:
                    for j in range(3):
                        if (j!=1):
                            self.hamiltonian_section+=(R'-1/2*'+self.massargs[i]).ljust(25)+R' |'
                        else:
                            self.hamiltonian_section+=(R'-1/8*'+self.massargs[i]).ljust(25)+R' |'
                        l=0
                        for k in range(len(self.massargs)): #check mass args
                            #print([l,self.masspos[i]])
                            if (l!=self.masspos[i][0]):
                                if 'e' in self.massargs[k]:
                                    self.hamiltonian_section+=(' 1').ljust(colsize[l])+R' |'
                                    l=l+1
                                    self.hamiltonian_section+=(' 1').ljust(colsize[l])+R' |'
                                    l=l+1
                                else:
                                    self.hamiltonian_section+=(' 1').ljust(colsize[l])+R' |'
                                    l=l+1
                            else:
                                if (j==0):
                                    self.hamiltonian_section+=' dq^2'.ljust(colsize[l])+R' |'
                                    l=l+1
                                    self.hamiltonian_section+=' 1'.ljust(colsize[l])+R' |'
                                    l=l+1
                                elif (j==1):
                                    self.hamiltonian_section+=' q^-2'.ljust(colsize[l])+R' |'
                                    l=l+1
                                    self.hamiltonian_section+=' 1'.ljust(colsize[l])+R' |'
                                    l=l+1
                                else:
                                    self.hamiltonian_section+=' q^-2'.ljust(colsize[l])+R' |'
                                    l=l+1
                                    self.hamiltonian_section+=' dq^2'.ljust(colsize[l])+R' |'
                                    l=l+1
                                
                        self.hamiltonian_section+=' 1'.ljust(colsize[k])+'\n'
            else:
                self.hamiltonian_section+=(R'1').ljust(25)+' |'
                for k in range(self.nmodes-1):
                    if (k!=self.masspos[i]):
                        self.hamiltonian_section+=(' 1').ljust(colsize[k])+R' |'
                    else:
                        self.hamiltonian_section+=(' KE').ljust(colsize[k])+R' |'
                self.hamiltonian_section+=(' 1').ljust(colsize[k])+'\n'
    
    def addpbasis(self):
        self.inpfile=self.inpfile+'PBASIS-SECTION'+'\n'
        for i in range(self.nmodes):
            if ('rho' in self.modeargs[i]):
                self.inpfile+=self.modeargs[i].ljust(8)+'lagu1'.ljust(8)+'35'.ljust(8)+' 0.0 0.5'+'\n'
            elif ('phi' in self.modeargs[i]):
                self.inpfile+=self.modeargs[i].ljust(8)+'exp'.ljust(8)+'35'.ljust(8)+' 0.00 6.283185307179586 periodic'+'\n'
            elif ('el' in self.modeargs[i]):
                self.inpfile+='el'.ljust(8)+'el'.ljust(8)+'4'.ljust(8)+'\n'
            else:
                self.inpfile+=self.modeargs[i].ljust(8)+'HO'.ljust(8)+'21'.ljust(8)+' 0.0 1.0 1.0'+'\n'
        self.inpfile+='end-pbasis-section'+'\n'


    def appendmctdh(self,sympy_str,pos,e_coords):
        #j=0 is treated differently
        js=0
        
        if (sympy_str[0]!='0'):
            vtt=sympy_str[0].split('*')
            #print('in append')
            #print(sympy_str)
            
            #print(vtt)
            #compose part before coefficient
        
            if (vtt[0][0]=='-'):
                firstcoef='-'
                vtt[0]=vtt[0][1:]
            else:
                firstcoef=None
        
        
            #print('first term')
            #print(vtt[0])
            js=self.startline(vtt,firstcoef)
            #print(vtt[js:])
            if (e_coords=='cart'):
                self.add_hamiltonian_line(vtt[js:],pos)
            else:
                self.add_hamiltonian_line_p(vtt[js:],pos)
            #print(self.hamiltonian_section)
            #exit()
        

        for j in range(1,len(sympy_str)):
            if (sympy_str[j]=='-'):
                firstcoef='-'
            elif (sympy_str[j]=='+'):
                firstcoef=None
            else:
                vtt=sympy_str[j].split('*')
                js=self.startline(vtt,firstcoef)
                #print(vtt[js:])
                if (e_coords=='cart'):
                    self.add_hamiltonian_line(vtt[js:],pos)
                else:
                    self.add_hamiltonian_line_p(vtt[js:],pos)
            
            
        #print(self.hamiltonian_section)

    def startline(self,vtt,firstcoef):
        js=0
        if (vtt[0][0].isdigit()):
            if firstcoef is None:
                if (vtt[0]!='1'):
                    firstcoef=vtt[0]+'*'
            else:
                if (vtt[0]!='1'):
                    firstcoef=firstcoef+vtt[0]+'*'
            js=js+1
        else:
            firstcoef==None
        if (vtt[js]=='sqrt(2)'):
            if (firstcoef is not None):
                firstcoef=firstcoef+'sq2'+'*'
            else:
                firstcoef='sq2'+'*'
            if ('sq2' not in self.parameters_list):
                self.parameters_list.append('sq2')
            js=js+1
        newparam=self.convert_param(vtt[js])
        if (newparam not in self.parameters_list):
            self.parameters_list.append(newparam)
        if (firstcoef is not None):
            firstcoef=firstcoef+newparam
        else:
            firstcoef=newparam
        
        #if vtt[0][0].isdigit():
        #    firstc=vtt[0]+'*'+param
        #elif vtt[0][0]=='-':
        #    if len(vtt[0])>1:
        #        firstc=vtt[0]+'*'+param
        #    else:
        #        firstc=vtt[0]+param
        #else:
        #    firstc=param
        self.hamiltonian_section+=firstcoef.ljust(25)+R' |'
        return js+1

    def add_hamiltonian_line(self,vt,pos):
        lenvt1=len(vt)
        for i in range(lenvt1):
            vt[i]=vt[i].replace('_','')
        #print('----------------')
        #print(self.nmodes)
        for i in range(self.nmodes-1):
            #print(i,self.modeargs[i],vt)
            if self.modeargs[i] in vt:
                mind=vt.index(self.modeargs[i])
                if mind<lenvt1-2:
                    if vt[mind+1]=='':
                        expo=int(vt[mind+2])
                    else:
                        expo=1
                else:
                    expo=1
                if expo>1:
                    self.hamiltonian_section+=(' q'+R'^'+f'{expo}').ljust(5)
                else:
                    self.hamiltonian_section+=(' q').ljust(5)
            else:
                self.hamiltonian_section+=(' 1').ljust(5)
            if i<self.nmodes-1:
                self.hamiltonian_section+=R' |'
        
        self.hamiltonian_section+=f' S{pos[0]}'
        self.hamiltonian_section+=R'&'
        self.hamiltonian_section+=f'{pos[1]}'
        self.hamiltonian_section+='\n'                
                #for i in range(vs,len(vt)):
                #    if vt[i][0].isalpha():
                #        mctdh=mctdh+'*'+vt[i]
                #    if vt[i][0]
    def convert_param(self,param):
        sparam=list(str(param))
        paramout=sparam[0].capitalize()
        for i in range(1,len(sparam)):
            if sparam[i].isalpha() or sparam[i].isdigit() or sparam[i]=='-':
                if sparam[i]=='-':
                    paramout=paramout+'m'

                else:
                    paramout=paramout+sparam[i]
        return paramout

    def finalize(self):
        self.hamiltonian_section+='end-hamiltonian-section'+'\n'
        self.hamiltonian_section+=''+'\n'
        for i in range(len(self.parameters_list)):
            if (self.parameters_list[i]=='sq2'):
                self.parameter_section+=(self.parameters_list[i]+'='+'1.414213562373095'+'\n')
            else:
                self.parameter_section+=(self.parameters_list[i]+'='+'\n')
        self.parameter_section+='end-parameter-section'+'\n'
        self.parameter_section+=''+'\n'
        self.fileform=self.title_section+self.parameter_section+self.hamiltonian_section+'end-operator'+'\n'

    def write_to_screen(self):
        output=self.title_section+self.parameter_section+self.hamiltonian_section+'end-operator'+'\n'
        print(output)

    def add_hamiltonian_line_p(self,vt,pos):
        lenvt1=len(vt)
        for i in range(lenvt1):
            vt[i]=vt[i].replace('_','')
        #print('----------------')
        #print(self.nmodes)
        #print(vt)
        #exit()
        for i in range(self.nmodes-1):
            #print(i,self.modeargs[i],vt)
            phitrue=False
            if ('phi' in self.modeargs[i]):
                phitrue=True
                colsize=12
                sinphi=R'sin('+self.modeargs[i]+R')'
                cosphi=R'cos('+self.modeargs[i]+R')'
            else:
                colsize=5
            if phitrue==False:
                if self.modeargs[i] in vt: 
                    mind=vt.index(self.modeargs[i])
                    if mind<lenvt1-2:
                        if vt[mind+1]=='':
                            expo=int(vt[mind+2])
                        else:
                            expo=1
                    else:
                        expo=1
                    if expo>1:
                        self.hamiltonian_section+=(' q'+R'^'+f'{expo}').ljust(colsize)
                    else:
                        self.hamiltonian_section+=(' q').ljust(colsize)
                else:
                    self.hamiltonian_section+=(' 1').ljust(colsize)
            if phitrue==True:
                #both sin(phin) and cos(phin) can be present
                if sinphi in vt or cosphi in vt:
                    hamadd=None
                    if (sinphi in vt):
                        mind=vt.index(sinphi)
                        if mind<lenvt1-2:
                            if vt[mind+1]=='':
                                expo=int(vt[mind+2])
                            else:
                                expo=1
                        else:
                            expo=1
                        if expo>1:
                            hamadd=(' sin'+R'^'+f'{expo}')
                        else:
                            hamadd=(' sin')
                    if (cosphi in vt):
                        mind=vt.index(cosphi)
                        if mind<lenvt1-2:
                            if vt[mind+1]=='':
                                expo=int(vt[mind+2])
                            else:
                                expo=1
                        else:
                            expo=1
                        if expo>1:
                            if hamadd is not None:
                                hamadd+='*'+('cos'+R'^'+f'{expo}')
                            else:
                                hamadd=(' cos'+R'^'+f'{expo}')
                        else:
                            if hamadd is not None:
                                hamadd+='*'+('cos')
                            else:
                                hamadd=(' cos')
                    self.hamiltonian_section+=hamadd.ljust(colsize)
                else:
                    self.hamiltonian_section+=(' 1').ljust(colsize)
            
            if i<self.nmodes-1:
                self.hamiltonian_section+=R' |'
        
        self.hamiltonian_section+=f' S{pos[0]}'
        self.hamiltonian_section+=R'&'
        self.hamiltonian_section+=f'{pos[1]}'
        self.hamiltonian_section+='\n'                
                #for i in range(vs,len(vt)):
                #    if vt[i][0].isalpha():
                #        mctdh=mctdh+'*'+vt[i]
                #    if vt[i][0]
#EOF
